# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains classes and methods to load input data for Meridian.

The `InputDataLoader` abstract class defines a single method: `load()` which
reads data from any of the supported sources and stores it as an `InputData`
object.
"""

import abc
from collections.abc import Mapping, Sequence
import dataclasses
import immutabledict
from meridian import constants
from meridian.data import data_frame_input_data_builder
from meridian.data import input_data
from meridian.data import input_data_builder
import pandas as pd
import xarray as xr


__all__ = [
    'InputDataLoader',
    'XrDatasetDataLoader',
    'DataFrameDataLoader',
    'CoordToColumns',
    'CsvDataLoader',
]


class InputDataLoader(metaclass=abc.ABCMeta):
  """Loads the data from the specified data format."""

  @abc.abstractmethod
  def load(self) -> input_data.InputData:
    """Reads the data and outputs an `InputData` object."""
    raise NotImplementedError()


class XrDatasetDataLoader(InputDataLoader):
  """Reads data from an `xarray.Dataset` object.

  Attributes:
    dataset: An `xr.Dataset` object containing the input data.
    kpi_type: A string denoting whether the KPI is of a `'revenue'` or
      `'non-revenue'` type. When the `kpi_type` is `'non-revenue'` and
      `revenue_per_kpi` exists, ROI calibration is used and the analysis is run
      on revenue. When `revenue_per_kpi` doesn't exist for the same `kpi_type`,
      custom ROI calibration is used and the analysis is run on KPI.

  Example:

  ```python
    data_loader = XrDatasetDataLoader(pickle.loads('data.pickle'))
    data = data_loader.load()
  ```
  """

  dataset: xr.Dataset

  def __init__(
      self,
      dataset: xr.Dataset,
      kpi_type: str,
      name_mapping: Mapping[str, str] | None = None,
  ):
    """Constructor.

    The coordinates of the input dataset should be: `time`, `media_time`,
    `control_variable` (optional), `geo` (optional for a national model),
    `non_media_channel` (optional), `organic_media_channel` (optional),
    `organic_rf_channel` (optional), and
    either `media_channel`, `rf_channel`, or both.

    Coordinate labels for `time` and `media_time` must be formatted in
    `"yyyy-mm-dd"` date format.

    In a geo model, the dataset should consist of the following arrays of the
    following dimensions. We use `1` to indicate a required dimension with
    length 1:

    *   `kpi`: `(geo, time)`
    *   `revenue_per_kpi`: `(geo, time)`
    *   `controls`: `(geo, time, control_variable)` - optional
    *   `population`: `(geo)`
    *   `media`: `(geo, media_time, media_channel)` - optional
    *   `media_spend`: `(geo, time, media_channel)`, `(1, time, media_channel)`,
        `(geo, 1, media_channel)`, `(media_channel)` - optional
    *   `reach`: `(geo, media_time, rf_channel)` - optional
    *   `frequency`: `(geo, media_time, rf_channel)` - optional
    *   `rf_spend`: `(geo, time, rf_channel)`, `(1, time, rf_channel)`,
        `(geo, 1, rf_channel)`, or `(rf_channel)` - optional
    *   `non_media_treatments`: `(geo, time, non_media_channel)` - optional
    *   `organic_media`: `(geo, media_time, organic_media_channel)` - optional
    *   `organic_reach`: `(geo, media_time, organic_rf_channel)` - optional
    *   `organic_frequency`: `(geo, media_time, organic_rf_channel)` - optional

    In a national model, the dataset should consist of the following arrays of
    the following dimensions. We use `[1,]` to indicate an optional dimension
    with length 1:

    *   `kpi`: `([1,] time)`
    *   `revenue_per_kpi`: `([1,] time)`
    *   `controls`: `([1,] time, control_variable)` - optional
    *   `population`: `([1],)` - this array is optional for national data
    *   `media`: `([1,] media_time, media_channel)` - optional
    *   `media_spend`: `([1,] time, media_channel)` or
        `([1,], [1,], media_channel)` - optional
    *   `reach`: `([1,] media_time, rf_channel)` - optional
    *   `frequency`: `([1,] media_time, rf_channel)` - optional
    *   `rf_spend`: `([1,] time, rf_channel)` or `([1,], [1,], rf_channel)` -
        optional
    *   `non_media_treatments`: `([1,] time, non_media_channel)` - optional
    *   `organic_media`: `([1,] media_time, organic_media_channel)` - optional
    *   `organic_reach`: `([1,] media_time, organic_rf_channel)` - optional
    *   `organic_frequency`: `([1,] media_time, organic_rf_channel)` - optional

    In a national model, the data will be expanded to include a single geo
    dimension.

    The dataset should include at least one of the following metric
    combinations: (1) media and media_spend or (2) reach, frequency, rf_spend.

    If the names of the coordinates or arrays are different, they can be renamed
    using the name_mapping argument. Example:

    ```python
    loader = XrDatasetDataLoader(
        dataset=pickle.loads('data.pickle'),
        name_mapping={'group': 'geo', 'cost': 'media_spend', 'conversions':
        'kpi'},
    )
    ```

    Alternatively to using `media_time`, the `media`, `reach` and `frequency`
    arrays can use the `time` coordinate, the same as the other arrays use.
    In such case, the dimensions will be converted by the loader into `time`
    and `media_time` and the lagged period will be determined by the missing
    values in the other arrays, similarly to `DataFrameDataLoader` and
    `CsvDataLoader`.

    Args:
      dataset: An `xarray.Dataset` object containing the input data.
      kpi_type: A string denoting whether the KPI is of a `'revenue'` or
        `'non-revenue'` type. When the `kpi_type` is `'non-revenue'` and
        `revenue_per_kpi` exists, ROI calibration is used and the analysis is
        run on revenue. When `revenue_per_kpi` doesn't exist for the same
        `kpi_type`, custom ROI calibration is used and the analysis is run on
        KPI.
      name_mapping: An optional dictionary whose keys are the current
        coordinates or array names in the `input` dataset and whose values are
        the desired coordinates (`geo`, `time`, `media_time`, `media_channel`
        and/or `rf_channel`, `control_variable`, `non_media_channel`,
        `organic_media_channel`, `organic_rf_channel`) or array names (`kpi`,
        `revenue_per_kpi`, `media`, `media_spend` and/or `rf_spend`, `controls`,
        `population`, `non_media_treatments`, `organic_media`, `organic_reach`,
        `organic_frequency`). Mapping must be provided if the names in the
        `input` dataset are different from the required ones, otherwise errors
        are thrown.
    """
    self.kpi_type = kpi_type
    if name_mapping is None:
      self.dataset = dataset
    else:
      source_coord_names = tuple(dataset.coords.keys())
      source_array_names = tuple(dataset.data_vars.keys())
      source_coords_and_arrays_set = frozenset(
          source_coord_names + source_array_names
      )
      for name in name_mapping.values():
        if name not in constants.POSSIBLE_INPUT_DATA_COORDS_AND_ARRAYS_SET:
          raise ValueError(
              f"Target name '{name}' from the mapping is none of the target"
              f' coordinate names {constants.POSSIBLE_INPUT_DATA_COORD_NAMES}'
              f' or array names {constants.POSSIBLE_INPUT_DATA_ARRAY_NAMES}.'
          )

      for name in name_mapping.keys():
        if name not in source_coords_and_arrays_set:
          raise ValueError(
              f"Source name '{name}' from the mapping is none of the"
              f' coordinate names {source_coord_names} or array names'
              f' {source_array_names} of the input dataset.'
          )

      self.dataset = dataset.rename(name_mapping)

    # Add a `geo` dimension if it is not already present.
    if (constants.GEO) not in self.dataset.sizes.keys():
      self.dataset = self.dataset.expand_dims(dim=[constants.GEO], axis=0)

    if constants.MEDIA_TIME not in self.dataset.sizes.keys():
      na_mask = self.dataset[constants.KPI].isnull().any(dim=constants.GEO)

      if constants.CONTROLS in self.dataset.data_vars.keys():
        na_mask |= (
            self.dataset[constants.CONTROLS]
            .isnull()
            .any(dim=[constants.GEO, constants.CONTROL_VARIABLE])
        )

      if constants.NON_MEDIA_TREATMENTS in self.dataset.data_vars.keys():
        na_mask |= (
            self.dataset[constants.NON_MEDIA_TREATMENTS]
            .isnull()
            .any(dim=[constants.GEO, constants.NON_MEDIA_CHANNEL])
        )

      if constants.REVENUE_PER_KPI in self.dataset.data_vars.keys():
        na_mask |= (
            self.dataset[constants.REVENUE_PER_KPI]
            .isnull()
            .any(dim=constants.GEO)
        )
      if constants.MEDIA_SPEND in self.dataset.data_vars.keys():
        na_mask |= (
            self.dataset[constants.MEDIA_SPEND]
            .isnull()
            .any(dim=[constants.GEO, constants.MEDIA_CHANNEL])
        )
      if constants.RF_SPEND in self.dataset.data_vars.keys():
        na_mask |= (
            self.dataset[constants.RF_SPEND]
            .isnull()
            .any(dim=[constants.GEO, constants.RF_CHANNEL])
        )

      # Dates with at least one non-NA value in non-media columns
      no_na_period = self.dataset[constants.TIME].isel(time=~na_mask).values

      # Create new `time` and `media_time` coordinates.
      new_time = 'new_time'

      new_dataset = self.dataset.assign_coords(
          new_time=(new_time, no_na_period),
      )

      new_dataset[constants.KPI] = (
          new_dataset[constants.KPI]
          .dropna(dim=constants.TIME)
          .rename({constants.TIME: new_time})
      )
      if constants.CONTROLS in new_dataset.data_vars.keys():
        new_dataset[constants.CONTROLS] = (
            new_dataset[constants.CONTROLS]
            .dropna(dim=constants.TIME)
            .rename({constants.TIME: new_time})
        )
      if constants.NON_MEDIA_TREATMENTS in new_dataset.data_vars.keys():
        new_dataset[constants.NON_MEDIA_TREATMENTS] = (
            new_dataset[constants.NON_MEDIA_TREATMENTS]
            .dropna(dim=constants.TIME)
            .rename({constants.TIME: new_time})
        )

      if constants.REVENUE_PER_KPI in new_dataset.data_vars.keys():
        new_dataset[constants.REVENUE_PER_KPI] = (
            new_dataset[constants.REVENUE_PER_KPI]
            .dropna(dim=constants.TIME)
            .rename({constants.TIME: new_time})
        )

      if constants.MEDIA_SPEND in new_dataset.data_vars.keys():
        new_dataset[constants.MEDIA_SPEND] = (
            new_dataset[constants.MEDIA_SPEND]
            .dropna(dim=constants.TIME)
            .rename({constants.TIME: new_time})
        )

      if constants.RF_SPEND in new_dataset.data_vars.keys():
        new_dataset[constants.RF_SPEND] = (
            new_dataset[constants.RF_SPEND]
            .dropna(dim=constants.TIME)
            .rename({constants.TIME: new_time})
        )

      self.dataset = new_dataset.rename(
          {constants.TIME: constants.MEDIA_TIME, new_time: constants.TIME}
      )

  def load(self) -> input_data.InputData:
    """Returns an `InputData` object containing the data from the dataset."""
    builder = input_data_builder.InputDataBuilder(self.kpi_type)
    builder.kpi = self.dataset.kpi
    if constants.POPULATION in self.dataset.data_vars.keys():
      builder.population = self.dataset.population
    if constants.CONTROLS in self.dataset.data_vars.keys():
      builder.controls = self.dataset.controls
    if constants.REVENUE_PER_KPI in self.dataset.data_vars.keys():
      builder.revenue_per_kpi = self.dataset.revenue_per_kpi
    if constants.MEDIA in self.dataset.data_vars.keys():
      builder.media = self.dataset.media
    if constants.MEDIA_SPEND in self.dataset.data_vars.keys():
      builder.media_spend = self.dataset.media_spend
    if constants.REACH in self.dataset.data_vars.keys():
      builder.reach = self.dataset.reach
    if constants.FREQUENCY in self.dataset.data_vars.keys():
      builder.frequency = self.dataset.frequency
    if constants.RF_SPEND in self.dataset.data_vars.keys():
      builder.rf_spend = self.dataset.rf_spend
    if constants.NON_MEDIA_TREATMENTS in self.dataset.data_vars.keys():
      builder.non_media_treatments = self.dataset.non_media_treatments
    if constants.ORGANIC_MEDIA in self.dataset.data_vars.keys():
      builder.organic_media = self.dataset.organic_media
    if constants.ORGANIC_REACH in self.dataset.data_vars.keys():
      builder.organic_reach = self.dataset.organic_reach
    if constants.ORGANIC_FREQUENCY in self.dataset.data_vars.keys():
      builder.organic_frequency = self.dataset.organic_frequency
    return builder.build()


@dataclasses.dataclass(frozen=True)
class CoordToColumns:
  """A mapping between the desired and actual column names in the input data.

  Attributes:
    time: Name of column containing `time` values in the input data.
    geo:  Name of column containing `geo` values in the input data. This field
      is optional for a national model.
    kpi: Name of column containing `kpi` values in the input data.
    controls: List of column names containing `controls` values in the input
      data. Optional.
    revenue_per_kpi: Name of column containing `revenue_per_kpi` values in the
      input data. Optional. Will be overridden if model KPI type is "revenue".
    population: Name of column containing `population` values in the input data.
      This field is optional for a national model.
    media: List of column names containing `media` values in the input data.
    media_spend: List of column names containing `media_spend` values in the
      input data.
    reach: List of column names containing `reach` values in the input data.
    frequency: List of column names containing `frequency` values in the input
      data.
    rf_spend: List of column names containing `rf_spend` values in the input
      data.
    non_media_treatments: List of column names containing `non_media_treatments`
      values in the input data.
    organic_media: List of column names containing `organic_media` values in the
      input data.
    organic_reach: List of column names containing `organic_reach` values in the
      input data.
    organic_frequency: List of column names containing `organic_frequency`
      values in the input data.
  """

  time: str = constants.TIME
  geo: str = constants.GEO
  kpi: str = constants.KPI
  controls: Sequence[str] | None = None
  revenue_per_kpi: str | None = None
  population: str = constants.POPULATION
  # Media data
  media: Sequence[str] | None = None
  media_spend: Sequence[str] | None = None
  # RF data
  reach: Sequence[str] | None = None
  frequency: Sequence[str] | None = None
  rf_spend: Sequence[str] | None = None
  # Non-media treatments data
  non_media_treatments: Sequence[str] | None = None
  # Organic media and RF data
  organic_media: Sequence[str] | None = None
  organic_reach: Sequence[str] | None = None
  organic_frequency: Sequence[str] | None = None

  def __post_init__(self):
    has_media_fields = self.media and self.media_spend
    has_rf_fields = self.reach and self.frequency and self.rf_spend
    if not (has_media_fields or has_rf_fields):
      raise ValueError(
          '`coord_to_columns` should include media data (`media` and'
          ' `media_spend`) or RF data (`reach`, `frequency` and `rf_spend`), or'
          ' both.'
      )

    if self.revenue_per_kpi is not None and not self.revenue_per_kpi.strip():
      raise ValueError('`revenue_per_kpi` should not be empty if provided.')


@dataclasses.dataclass
class DataFrameDataLoader(InputDataLoader):
  """Reads data from a Pandas `DataFrame`.

  This class reads input data from a Pandas `DataFrame`. The `coord_to_columns`
  attribute stores a mapping from target `InputData` coordinates and array names
  to the DataFrame column names if they are different. The fields are:

  *   `geo`, `time`, `kpi`, `revenue_per_kpi`, `population` (single column)
  *   `controls` (multiple columns, optional)
  *   (1) `media`, `media_spend` (multiple columns)
  *   (2) `reach`, `frequency`, `rf_spend` (multiple columns)
  *   `non_media_treatments` (multiple columns, optional)
  *   `organic_media` (multiple columns, optional)
  *   `organic_reach`, `organic_frequency` (multiple columns, optional)

  The `DataFrame` must include (1) or (2), but doesn't need to include both.
  Also, each media channel must appear in (1) or (2), but not both.

  Note the following:

  *   Time column values must be formatted in _yyyy-mm-dd_ date format.
  *   In a national model, `geo` and `population` are optional. If the
      `population` is provided, it is reset to a default value of `1.0`.
  *   If `media` data is provided, then `media_to_channel` and
      `media_spend_to_channel` are required. If `reach` and `frequency` data is
      provided, then `reach_to_channel` and `frequency_to_channel` and
      `rf_spend_to_channel` are required.
  *   If `organic_reach` and `organic_frequency` data is provided, then
      `organic_reach_to_channel` and `organic_frequency_to_channel` are
      required.

  Example:

  ```python
  # df = [...]
  coord_to_columns = CoordToColumns(
    geo='dmas',
    time='dates',
    kpi='conversions',
    revenue_per_kpi='revenue_per_conversions',
    controls=['control_income'],
    population='populations',
    media=['impressions_tv', 'impressions_fb', 'impressions_search'],
    media_spend=['spend_tv', 'spend_fb', 'spend_search'],
    reach=['reach_yt'],
    frequency=['frequency_yt'],
    rf_spend=['rf_spend_yt'],
    non_media_treatments=['price', 'discount']
    organic_media=['organic_impressions_blog'],
    organic_reach=['organic_reach_newsletter'],
    organic_frequency=['organic_frequency_newsletter'],
  )
  media_to_channel = {
      'impressions_tv': 'tv',
      'impressions_fb': 'fb',
      'impressions_search': 'search',
  }
  media_spend_to_channel = {
      'spend_tv': 'tv', 'spend_fb': 'fb', 'spend_search': 'search'
  }
  reach_to_channel = {'reach_yt': 'yt'}
  frequency_to_channel = {'frequency_yt': 'yt'}
  rf_spend_to_channel = {'rf_spend_yt': 'yt'}
  organic_reach_to_channel = {'organic_reach_newsletter': 'newsletter'}
  organic_frequency_to_channel = {'organic_frequency_newsletter': 'newsletter'}

  data_loader = DataFrameDataLoader(
      df=df,
      coord_to_columns=coord_to_columns,
      kpi_type='non-revenue',
      media_to_channel=media_to_channel,
      media_spend_to_channel=media_spend_to_channel,
      reach_to_channel=reach_to_channel,
      frequency_to_channel=frequency_to_channel,
      rf_spend_to_channel=rf_spend_to_channel,
      organic_reach_to_channel=organic_reach_to_channel,
      organic_frequency_to_channel=organic_frequency_to_channel,
  )
  data = data_loader.load()
  ```

  Attributes:
    df: The `pd.DataFrame` object to read from. One of the following conditions
      is required:

      *   There are no NAs in the dataframe
      *   For any number of initial periods there is only media data and NAs in
          all of the non-media data columns (`kpi`, `revenue_per_kpi`,
          `media_spend`, `controls`, and `population`).

    coord_to_columns: A `CoordToColumns` object whose fields are the desired
      coordinates of the `InputData` and the values are the current names of
      columns (or lists of columns) in the DataFrame. Example:

      ```
      coord_to_columns = CoordToColumns(
          geo='dmas',
          time='dates',
          kpi='conversions',
          revenue_per_kpi='revenue_per_conversions',
          media=['impressions_tv', 'impressions_yt', 'impressions_search'],
          spend=['spend_tv', 'spend_yt', 'spend_search'],
          controls=['control_income'],
          population=population,
      )
      ```

    kpi_type: A string denoting whether the KPI is of a `'revenue'` or
      `'non-revenue'` type. When the `kpi_type` is `'non-revenue'` and there
      exists a `revenue_per_kpi`, ROI calibration is used and the analysis is
      run on revenue. When the `revenue_per_kpi` doesn't exist for the same
      `kpi_type`, custom ROI calibration is used and the analysis is run on KPI.
    media_to_channel: A dictionary whose keys are the actual column names for
      `media` data in the dataframe, and the values are the desired channel
      names. These are the same as for the `media_spend` data. Example:

      ```
      media_to_channel = {'media_tv': 'tv', 'media_yt': 'yt', 'media_fb': 'fb'}
      ```

    media_spend_to_channel: A dictionary whose keys are the actual column names
      for `media_spend` data in the dataframe, and the values are the desired
      channel names. These are same as for the `media` data. Example:

      ```
      media_spend_to_channel = {
          'spend_tv': 'tv', 'spend_yt': 'yt', 'spend_fb': 'fb'
      }
      ```

    reach_to_channel: A dictionary whose keys are the actual column names for
      `reach` data in the dataframe, and the values are the desired channel
      names. These are the same as for the `rf_spend` data. Example:

      ```
      reach_to_channel = {'reach_tv': 'tv', 'reach_yt': 'yt', 'reach_fb': 'fb'}
      ```

    frequency_to_channel: A dictionary whose keys are the actual column names
      for `frequency` data in the dataframe, and the values are the desired
      channel names. These are the same as for the `rf_spend` data. Example:

      ```
      frequency_to_channel = {
          'frequency_tv': 'tv', 'frequency_yt': 'yt', 'frequency_fb': 'fb'
      }
      ```

    rf_spend_to_channel: A dictionary whose keys are the actual column names for
      `rf_spend` data in the dataframe, and values are the desired channel
      names. These are the same as for the `reach` and `frequency` data.
      Example:

      ```
      rf_spend_to_channel = {
          'rf_spend_tv': 'tv', 'rf_spend_yt': 'yt', 'rf_spend_fb': 'fb'
      }
      ```

    organic_reach_to_channel: A dictionary whose keys are the actual column names
      for `organic_reach` data in the dataframe, and the values are the desired
      channel names. These are the same as for the `organic_frequency` data.
      Example:

      ```
      organic_reach_to_channel = {
          'organic_reach_newsletter': 'newsletter',
      }
      ```

    organic_frequency_to_channel: A dictionary whose keys are the actual column
      names for `organic_frequency` data in the dataframe, and the values are
      the desired channel names. These are the same as for the `organic_reach`
      data. Example:

      ```
      organic_frequency_to_channel = {
          'organic_frequency_newsletter': 'newsletter',
      }
      ```
  """  # pyformat: disable

  df: pd.DataFrame
  coord_to_columns: CoordToColumns
  kpi_type: str
  media_to_channel: Mapping[str, str] | None = None
  media_spend_to_channel: Mapping[str, str] | None = None
  reach_to_channel: Mapping[str, str] | None = None
  frequency_to_channel: Mapping[str, str] | None = None
  rf_spend_to_channel: Mapping[str, str] | None = None
  organic_reach_to_channel: Mapping[str, str] | None = None
  organic_frequency_to_channel: Mapping[str, str] | None = None

  def __post_init__(self):
    # If [key] in the following dict exists as an attribute in
    # `coord_to_columns`, then the corresponding attribute must exist in this
    # loader instance.
    required_mappings = immutabledict.immutabledict({
        'media': 'media_to_channel',
        'media_spend': 'media_spend_to_channel',
        'reach': 'reach_to_channel',
        'frequency': 'frequency_to_channel',
        'rf_spend': 'rf_spend_to_channel',
        'organic_reach': 'organic_reach_to_channel',
        'organic_frequency': 'organic_frequency_to_channel',
    })
    for coord_name, channel_dict in required_mappings.items():
      if getattr(self.coord_to_columns, coord_name, None) is not None:
        if getattr(self, channel_dict, None) is None:
          raise ValueError(
              f"When {coord_name} data is provided, '{channel_dict}' is"
              ' required.'
          )
        else:
          if set(getattr(self, channel_dict)) != set(
              getattr(self.coord_to_columns, coord_name)
          ):
            raise ValueError(
                f'The {channel_dict} keys must have the same set of values as'
                f' the {coord_name} columns.'
            )

    if (
        self.media_to_channel is not None
        and self.media_spend_to_channel is not None
    ):
      if set(self.media_to_channel.values()) != set(
          self.media_spend_to_channel.values()
      ):
        raise ValueError(
            'The media and media_spend columns must have the same set of'
            ' channels.'
        )

      # The columns listed in `media` and `media_spend` must correspond to the
      # same channels, in user-given order!
      # For example, this is invalid:
      #   media = ['impressions_tv', 'impressions_yt']
      #   media_spend = ['spend_yt', 'spend_tv']
      # But we can only detect this after we map each `media` and `media_spend`
      # column to its canonical channel name.
      media_channels = [
          self.media_to_channel[c] for c in self.coord_to_columns.media
      ]
      media_spend_channels = [
          self.media_spend_to_channel[c]
          for c in self.coord_to_columns.media_spend
      ]
      if media_channels != media_spend_channels:
        raise ValueError(
            'The `media` and `media_spend` columns must correspond to the same'
            ' channels, in user order.'
        )

    if (
        self.reach_to_channel is not None
        and self.frequency_to_channel is not None
        and self.rf_spend_to_channel is not None
    ):
      if (
          set(self.reach_to_channel.values())
          != set(self.frequency_to_channel.values())
          != set(self.rf_spend_to_channel.values())
      ):
        raise ValueError(
            'The reach, frequency, and rf_spend columns must have the same set'
            ' of channels.'
        )

      # Same channel ordering concerns as for `media` and `media_spend`.
      reach_channels = [
          self.reach_to_channel[c] for c in self.coord_to_columns.reach
      ]
      frequency_channels = [
          self.frequency_to_channel[c] for c in self.coord_to_columns.frequency
      ]
      rf_spend_channels = [
          self.rf_spend_to_channel[c] for c in self.coord_to_columns.rf_spend
      ]
      if not (reach_channels == frequency_channels == rf_spend_channels):
        raise ValueError(
            'The `reach`, `frequency`, and `rf_spend` columns must correspond'
            ' to the same channels, in user order.'
        )

    if (
        self.organic_reach_to_channel is not None
        and self.organic_frequency_to_channel is not None
    ):
      if set(self.organic_reach_to_channel.values()) != set(
          self.organic_frequency_to_channel.values()
      ):
        raise ValueError(
            'The organic_reach and organic_frequency columns must have the'
            ' same set of channels.'
        )

      # Same channel ordering concerns as for `media` and `media_spend`.
      organic_reach_channels = [
          self.organic_reach_to_channel[c]
          for c in self.coord_to_columns.organic_reach
      ]
      organic_frequency_channels = [
          self.organic_frequency_to_channel[c]
          for c in self.coord_to_columns.organic_frequency
      ]
      if organic_reach_channels != organic_frequency_channels:
        raise ValueError(
            'The `organic_reach` and `organic_frequency` columns must'
            ' correspond to the same channels, in user order.'
        )

  def load(self) -> input_data.InputData:
    """Reads data from a dataframe and returns an InputData object."""

    builder = data_frame_input_data_builder.DataFrameInputDataBuilder(
        kpi_type=self.kpi_type
    ).with_kpi(
        self.df,
        self.coord_to_columns.kpi,
        self.coord_to_columns.time,
        self.coord_to_columns.geo,
    )

    if self.coord_to_columns.population in self.df.columns:
      builder.with_population(
          self.df, self.coord_to_columns.population, self.coord_to_columns.geo
      )

    if self.coord_to_columns.controls:
      builder.with_controls(
          self.df,
          list(self.coord_to_columns.controls),
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if self.coord_to_columns.non_media_treatments:
      builder.with_non_media_treatments(
          self.df,
          list(self.coord_to_columns.non_media_treatments),
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if self.coord_to_columns.revenue_per_kpi:
      builder.with_revenue_per_kpi(
          self.df,
          self.coord_to_columns.revenue_per_kpi,
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if (
        self.media_to_channel is not None
        and self.media_spend_to_channel is not None
    ):
      # Based on the invariant rule enforced in `__post_init__`, the columns
      # listed in `media` and `media_spend` are already validated to correspond
      # to the same channels, in user-given order.
      media_execution_columns = list(self.coord_to_columns.media)
      media_spend_columns = list(self.coord_to_columns.media_spend)
      # So now we can use one of the channel mapper dicts to get the canonical
      # channel names for each column.
      media_channel_names = [
          self.media_to_channel[c] for c in self.coord_to_columns.media
      ]
      builder.with_media(
          self.df,
          media_execution_columns,
          media_spend_columns,
          media_channel_names,
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if (
        self.reach_to_channel is not None
        and self.frequency_to_channel is not None
        and self.rf_spend_to_channel is not None
    ):
      # Based on the invariant rule enforced in `__post_init__`, the columns
      # listed in `reach`, `frequency`, and `rf_spend` are already validated
      # to correspond to the same channels, in user-given order.
      reach_columns = list(self.coord_to_columns.reach)
      frequency_columns = list(self.coord_to_columns.frequency)
      rf_spend_columns = list(self.coord_to_columns.rf_spend)
      # So now we can use one of the channel mapper dicts to get the canonical
      # channel names for each column.
      rf_channel_names = [
          self.reach_to_channel[c] for c in self.coord_to_columns.reach
      ]
      builder.with_reach(
          self.df,
          reach_columns,
          frequency_columns,
          rf_spend_columns,
          rf_channel_names,
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if self.coord_to_columns.organic_media:
      builder.with_organic_media(
          self.df,
          list(self.coord_to_columns.organic_media),
          list(self.coord_to_columns.organic_media),
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    if (
        self.organic_reach_to_channel is not None
        and self.organic_frequency_to_channel is not None
    ):
      # Based on the invariant rule enforced in `__post_init__`, the columns
      # listed in `organic_reach` and `organic_frequency` are already
      # validated to correspond to the same channels, in user-given order.
      organic_reach_columns = list(self.coord_to_columns.organic_reach)
      organic_frequency_columns = list(self.coord_to_columns.organic_frequency)
      # So now we can use one of the channel mapper dicts to get the canonical
      # channel names for each column.
      organic_rf_channel_names = [
          self.organic_reach_to_channel[c]
          for c in self.coord_to_columns.organic_reach
      ]
      builder.with_organic_reach(
          self.df,
          organic_reach_columns,
          organic_frequency_columns,
          organic_rf_channel_names,
          self.coord_to_columns.time,
          self.coord_to_columns.geo,
      )

    return builder.build()


class CsvDataLoader(InputDataLoader):
  """Reads data from a CSV file.

  This class reads input data from a CSV file. The `coord_to_columns` attribute
  stores a mapping from target `InputData` coordinates and array names to the
  CSV column names, if they are different. The fields are:

  *   `geo`, `time`, `kpi`, `revenue_per_kpi`, `population` (single column)
  *   `controls` (multiple columns, optional)
  *   (1) `media`, `media_spend` (multiple columns)
  *   (2) `reach`, `frequency`, `rf_spend` (multiple columns)
  *   `non_media_treatments` (multiple columns, optional)
  *   `organic_media` (multiple columns, optional)
  *   `organic_reach`, `organic_frequency` (multiple columns, optional)

  The DataFrame must include either (1) or (2), but doesn't need to include
  both.

  Note: Time column values must be formatted using the _yyyy-mm-dd_ date format.

  Internally, this class reads the CSV file into a Pandas DataFrame and then
  loads the data using `DataFrameDataLoader`.

  Note: In a national model, `geo` and `population` are optional. If
  `population` is provided, it is reset to a default value of `1.0`.
  """

  def __init__(
      self,
      csv_path: str,
      coord_to_columns: CoordToColumns,
      kpi_type: str,
      media_to_channel: Mapping[str, str] | None = None,
      media_spend_to_channel: Mapping[str, str] | None = None,
      reach_to_channel: Mapping[str, str] | None = None,
      frequency_to_channel: Mapping[str, str] | None = None,
      rf_spend_to_channel: Mapping[str, str] | None = None,
      organic_reach_to_channel: Mapping[str, str] | None = None,
      organic_frequency_to_channel: Mapping[str, str] | None = None,
  ):
    """Constructor.

    Reads CSV file into a Pandas DataFrame and uses it to create a
    `DataFrameDataLoader`.

    Args:
      csv_path: The path to the CSV file to read from. One of the following
        conditions is required:

        *   There are no gaps in the data.
        *   For up to `max_lag` initial periods there is only media data and
            empty cells in all the data columns different from `media`, `reach`,
            `frequency`, `organic_media`, `organic_reach` and
            `organic_frequency` (`kpi`, `revenue_per_kpi`, `media_spend`,
            `rf_spend`, `controls`, `population` and `non_media_treatments`).

      coord_to_columns: A `CoordToColumns` object whose fields are the desired
        coordinates of the `InputData` and the values are the current names of
        columns (or lists of columns) in the CSV file. Example:

        ```
        coord_to_columns = CoordToColumns(
            geo='dmas',
            time='dates',
            kpi='revenue',
            revenue_per_kpi='revenue_per_conversions',
            media=['impressions_tv', impressions_yt', 'impressions_search'],
            spend=['spend_tv', 'spend_yt', 'spend_search'],
            reach=['reach_fb'],
            frequency=['frequency_fb'],
            rf_spend=['rf_spend_fb'],
            controls=['control_income'],
            population='population',
            non_media_treatments=['price', 'discount'],
            organic_media=['organic_impressions_blog'],
            organic_reach=['organic_reach_newsletter'],
            organic_frequency=['organic_frequency_newsletter'],
        )
        ```

      kpi_type: A string denoting whether the KPI is of a `'revenue'` or
        `'non-revenue'` type. When the `kpi_type` is `'non-revenue'` and there
        exists a `revenue_per_kpi`, ROI calibration is used and the analysis is
        run on revenue. When the `revenue_per_kpi` doesn't exist for the same
        `kpi_type`, custom ROI calibration is used and the analysis is run on
        KPI.
      media_to_channel: A dictionary whose keys are the actual column names for
        media data in the CSV file and values are the desired channel names, the
        same as for the `media_spend` data. Example:

        ```
        media_to_channel = {
            'media_tv': 'tv', 'media_yt': 'yt', 'media_fb': 'fb'
        }
        ```

      media_spend_to_channel: A dictionary whose keys are the actual column
        names for `media_spend` data in the CSV file and values are the desired
        channel names, the same as for the `media` data. Example:

        ```
        `media_spend_to_channel = {
            'spend_tv': 'tv', 'spend_yt': 'yt', 'spend_fb': 'fb'
        }
        ```

      reach_to_channel: A dictionary whose keys are the actual column names for
        `reach` data in the dataframe and values are the desired channel names,
        the same as for the `rf_spend` data. Example:

        ```
        reach_to_channel = {
            'reach_tv': 'tv', 'reach_yt': 'yt', 'reach_fb': 'fb'
        }
        ```

      frequency_to_channel: A dictionary whose keys are the actual column names
        for `frequency` data in the dataframe and values are the desired channel
        names, the same as for the `rf_spend` data. Example:

        ```
        frequency_to_channel = {
            'frequency_tv': 'tv', 'frequency_yt': 'yt', 'frequency_fb': 'fb'
        }
        ```

      rf_spend_to_channel: A dictionary whose keys are the actual column names
        for `rf_spend` data in the dataframe and values are the desired channel
        names, the same as for the `reach` and `frequency` data. Example:

        ```
        rf_spend_to_channel = {
            'rf_spend_tv': 'tv', 'rf_spend_yt': 'yt', 'rf_spend_fb': 'fb'
        }
        ```

      organic_reach_to_channel: A dictionary whose keys are the actual column
        names for `organic_reach` data in the dataframe and values are the
        desired channel names, the same as for the `organic_frequency`. Example:

        ```
        organic_reach_to_channel = {
            'organic_reach_newsletter': 'newsletter',
        }
        ```

      organic_frequency_to_channel: A dictionary whose keys are the actual
        column names for `organic_frequency` data in the dataframe and values
        are the desired channel names, the same as for the `organic_reach`
        data. Example:

        ```
        organic_frequency_to_channel = {
            'organic_frequency_newsletter': 'newsletter',
        }
        ```

    Note: In a national model, `geo` and `population` are optional. If
    `population` is provided, it is reset to a default value of `1.0`.

    Note: If `media` data is provided, then `media_to_channel` and
    `media_spend_to_channel` are required. If `reach` and `frequency` data is
    provided, then `reach_to_channel`, `frequency_to_channel`, and
    `rf_spend_to_channel` are required.
    """  # pyformat: disable
    df = pd.read_csv(csv_path)
    self._df_loader = DataFrameDataLoader(
        df=df,
        coord_to_columns=coord_to_columns,
        kpi_type=kpi_type,
        media_to_channel=media_to_channel,
        media_spend_to_channel=media_spend_to_channel,
        reach_to_channel=reach_to_channel,
        frequency_to_channel=frequency_to_channel,
        rf_spend_to_channel=rf_spend_to_channel,
        organic_reach_to_channel=organic_reach_to_channel,
        organic_frequency_to_channel=organic_frequency_to_channel,
    )

  def load(self) -> input_data.InputData:
    """Reads data from a CSV file and returns an `InputData` object."""

    return self._df_loader.load()



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of `Distribution` objects for priors."""

from __future__ import annotations

import inspect
import types
from typing import Any, Sequence, TypeVar
import warnings

from meridian import backend
from meridian import constants
from meridian.model import prior_distribution as pd
from mmm.v1.model.meridian import meridian_model_pb2 as meridian_pb
from schema.serde import constants as sc
from schema.serde import function_registry as function_registry_utils
from schema.serde import serde

from tensorflow.core.framework import tensor_shape_pb2  # pylint: disable=g-direct-tensorflow-import

FunctionRegistry = function_registry_utils.FunctionRegistry


MeridianPriorDistributions = (
    meridian_pb.PriorTfpDistributions
)

_FUNCTION_REGISTRY_NAME = "function_registry"


# TODO: b/436637084 - Delete enumerated schema.
class DistributionSerde(
    serde.Serde[MeridianPriorDistributions, pd.PriorDistribution]
):
  """Serializes and deserializes a Meridian prior distributions container into a `Distribution` proto."""

  def __init__(self, function_registry: FunctionRegistry):
    """Initializes a `DistributionSerde` instance.

    Args:
      function_registry: A lookup table containing custom functions used by
        various `backend.tfd` classes. It's recommended to explicitly define the
        custom functions instead of using lambdas, as lambda functions may not
        be registered successfully.
    """
    self._function_registry = function_registry

  @property
  def function_registry(self) -> FunctionRegistry:
    return self._function_registry

  def serialize(
      self, obj: pd.PriorDistribution
  ) -> meridian_pb.PriorTfpDistributions:
    """Serializes the given Meridian priors container into a `MeridianPriorDistributions` proto."""
    proto = meridian_pb.PriorTfpDistributions()
    for param in constants.ALL_PRIOR_DISTRIBUTION_PARAMETERS:
      if not hasattr(obj, param):
        continue
      getattr(proto, param).CopyFrom(
          self._to_distribution_proto(getattr(obj, param))
      )
    proto.function_registry.update(self.function_registry.hashed_registry)
    return proto

  def deserialize(
      self,
      serialized: MeridianPriorDistributions,
      serialized_version: str = "",
      force_deserialization: bool = False,
  ) -> pd.PriorDistribution:
    """Deserializes the `PriorTfpDistributions` proto.

    WARNING: If any custom functions in the function registry are modified after
    serialization, the deserialized model can differ from the original model, as
    the original function's behavior is no longer guaranteed. This will result
    in an error during deserialization.

    For users who are intentionally changing functions and are confident that
    the changes will not affect the deserialized model, you can bypass safety
    mechanisms to force deserialization. See example:

    Args:
      serialized: A serialized `PriorDistributions` object.
      serialized_version: The version of the serialized Meridian model. This is
        used to handle changes in deserialization logic across different
        versions.
      force_deserialization: If True, bypasses the safety check that validates
        whether functions within `function_registry` have changed after
        serialization. Use with caution. This should only be used if you have
        intentionally modified a custom function and are confident that the
        changes will not affect the deserialized model. A safer alternative is
        to first deserialize the model with the original functions and then
        serialize it with the new ones.

    Returns:
      A deserialized `PriorDistribution` object.
    """
    kwargs = {}
    for param in constants.ALL_PRIOR_DISTRIBUTION_PARAMETERS:
      if not hasattr(serialized, param):
        continue
      # A parameter may be unspecified in a serialized proto message because:
      # (1) It is left unset for Meridian to set its default value.
      # (2) The message was created from a previous Meridian version after
      #     introducing a new parameter.
      if not serialized.HasField(param):
        continue
      param_name = getattr(serialized, param)
      if isinstance(serialized, meridian_pb.PriorTfpDistributions):
        if force_deserialization:
          warnings.warn(
              "You're attempting to deserialize a model while ignoring changes"
              " to custom functions. This is a risky operation that can"
              " potentially lead to a deserialized model that behaves"
              " differently from the original, resulting in unexpected behavior"
              " or model failure. We strongly recommend a safer two-step"
              " process: deserialize the model using the original function"
              " registry and reserialize the model using the updated registry."
              " Please proceed with caution."
          )
        else:
          stored_hashed_function_registry = getattr(
              serialized, _FUNCTION_REGISTRY_NAME
          )
          try:
            self.function_registry.validate(stored_hashed_function_registry)
          except ValueError as e:
            raise ValueError(
                f"An issue found during deserializing Distribution: {e}"
            ) from e

        kwargs[param] = self._from_distribution_proto(param_name)
      # copybara: strip_begin(legacy proto)
      elif isinstance(serialized, meridian_pb.PriorDistributions):
        kwargs[param] = _from_legacy_distribution_proto(param_name)
      # copybara: strip_end
    return pd.PriorDistribution(**kwargs)

  def _to_distribution_proto(
      self,
      dist: backend.tfd.Distribution,
  ) -> meridian_pb.TfpDistribution:
    """Converts a TensorFlow `Distribution` object to a `TfpDistribution` proto."""
    dist_name = type(dist).__name__
    dist_class = getattr(backend.tfd, dist_name)
    return meridian_pb.TfpDistribution(
        distribution_type=dist_name,
        parameters={
            name: self._to_parameter_value_proto(name, value, dist_class)
            for name, value in dist.parameters.items()
        },
    )

  def _to_bijector_proto(
      self,
      bijector: backend.bijectors.Bijector,
  ) -> meridian_pb.TfpBijector:
    """Converts a TensorFlow `Bijector` object to a `TfpBijector` proto."""
    bij_name = type(bijector).__name__
    bij_class = getattr(backend.bijectors, bij_name)
    return meridian_pb.TfpBijector(
        bijector_type=bij_name,
        parameters={
            name: self._to_parameter_value_proto(name, value, bij_class)
            for name, value in bijector.parameters.items()
        },
    )

  def _to_parameter_value_proto(
      self,
      param_name: str,
      value: Any,
      dist: backend.tfd.Distribution | backend.bijectors.Bijector,
  ) -> meridian_pb.TfpParameterValue:
    """Converts a TensorFlow `Distribution` parameter value to a `TfpParameterValue` proto."""
    # Handle built-in types.
    match value:
      case float():
        return meridian_pb.TfpParameterValue(scalar_value=value)
      case int():
        return meridian_pb.TfpParameterValue(int_value=value)
      case bool():
        return meridian_pb.TfpParameterValue(bool_value=value)
      case str():
        return meridian_pb.TfpParameterValue(string_value=value)
      case None:
        return meridian_pb.TfpParameterValue(none_value=True)
      case list():
        value_generator = (
            self._to_parameter_value_proto(param_name, v, dist) for v in value
        )
        return meridian_pb.TfpParameterValue(
            list_value=meridian_pb.TfpParameterValue.List(
                values=value_generator
            )
        )
      case dict():
        dict_value = {
            k: self._to_parameter_value_proto(param_name, v, dist)
            for k, v in value.items()
        }
        return meridian_pb.TfpParameterValue(
            dict_value=meridian_pb.TfpParameterValue.Dict(value_map=dict_value)
        )
      case backend.Tensor():
        return meridian_pb.TfpParameterValue(
            tensor_value=backend.make_tensor_proto(value)
        )
      case backend.tfd.Distribution():
        return meridian_pb.TfpParameterValue(
            distribution_value=self._to_distribution_proto(value)
        )
      case backend.bijectors.Bijector():
        return meridian_pb.TfpParameterValue(
            bijector_value=self._to_bijector_proto(value)
        )
      case backend.tfd.ReparameterizationType():
        fully_reparameterized = value == backend.tfd.FULLY_REPARAMETERIZED
        return meridian_pb.TfpParameterValue(
            fully_reparameterized=fully_reparameterized
        )
      case types.FunctionType():
        # Check for default value
        signature = inspect.signature(dist.__init__)
        param = signature.parameters[param_name]
        if param.default and param.default is value:
          return meridian_pb.TfpParameterValue(
              function_param=meridian_pb.TfpParameterValue.FunctionParam(
                  uses_default=True
              )
          )

        # Check against registry.
        function_key = self.function_registry.get_function_key(value)
        if function_key is not None:
          return meridian_pb.TfpParameterValue(
              function_param=meridian_pb.TfpParameterValue.FunctionParam(
                  function_key=function_key
              )
          )
        raise ValueError(
            f"Custom function `{param_name}` detected for"
            f" {type(dist).__name__}, but not found in registry. Please"
            " add custom functions to registry when saving models."
        )

    # Handle unsupported types.
    raise TypeError(f"Unsupported type: {type(value)}, {value}")

  def _from_distribution_proto(
      self,
      dist_proto: meridian_pb.TfpDistribution,
  ) -> backend.tfd.Distribution:
    """Converts a `Distribution` proto to a TensorFlow `Distribution` object."""
    dist_class_name = dist_proto.distribution_type
    dist_class = getattr(backend.tfd, dist_class_name)
    dist_parameters = dist_proto.parameters
    input_parameters = {
        k: self._unpack_tfp_parameters(k, v, dist_class)
        for k, v in dist_parameters.items()
    }
    return dist_class(**input_parameters)

  def _from_bijector_proto(
      self,
      dist_proto: meridian_pb.TfpBijector,
  ) -> backend.bijectors.Bijector:
    """Converts a `Bijector` proto to a TensorFlow `Bijector` object."""
    dist_class_name = dist_proto.bijector_type
    dist_class = getattr(backend.bijectors, dist_class_name)
    dist_parameters = dist_proto.parameters
    input_parameters = {
        name: self._unpack_tfp_parameters(name, value, dist_class)
        for name, value in dist_parameters.items()
    }

    return dist_class(**input_parameters)

  def _unpack_tfp_parameters(
      self,
      param_name: str,
      param_value: meridian_pb.TfpParameterValue,
      dist_class: backend.tfd.Distribution,
  ) -> Any:
    """Unpacks a `TfpParameterValue` proto into a Python value."""
    match param_value.WhichOneof("value_type"):
      # Handle built-in types.
      case "scalar_value":
        return param_value.scalar_value
      case "int_value":
        return param_value.int_value
      case "bool_value":
        return param_value.bool_value
      case "string_value":
        return param_value.string_value
      case "none_value":
        return None
      case "list_value":
        return [
            self._unpack_tfp_parameters(param_name, v, dist_class)
            for v in param_value.list_value.values
        ]
      case "dict_value":
        items = param_value.dict_value.value_map.items()
        return {
            key: self._unpack_tfp_parameters(key, value, dist_class)
            for key, value in items
        }

      # Handle custom types.
      case "tensor_value":
        return backend.to_tensor(backend.make_ndarray(param_value.tensor_value))
      case "distribution_value":
        return self._from_distribution_proto(param_value.distribution_value)
      case "bijector_value":
        return self._from_bijector_proto(param_value.bijector_value)
      case "fully_reparameterized":
        if param_value.fully_reparameterized:
          return backend.tfd.FULLY_REPARAMETERIZED
        else:
          return backend.tfd.NOT_FULLY_REPARAMETERIZED

      # Handle functions.
      case "function_param":
        function_param = param_value.function_param
        # Check against registry.
        if function_param.HasField("function_key"):
          registry = self.function_registry
          if function_param.function_key in registry:
            return registry.get(function_param.function_key)
        # Check for default value.
        if (
            function_param.HasField("uses_default")
            and function_param.uses_default
        ):
          signature = inspect.signature(dist_class.__init__)
          return signature.parameters[param_name].default
        raise ValueError(f"No function found for {param_name}")

      # Handle unsupported types.
      case _:
        raise ValueError(
            f"Unsupported TFP distribution parameter type: {type(param_value)}"
        )

# copybara: strip_begin


def _from_legacy_bijector_proto(
    bijector_proto: meridian_pb.Distribution.Bijector,
) -> backend.bijectors.Bijector:
  """Converts a `Bijector` proto to a `Bijector` object."""
  bijector_type_field = bijector_proto.WhichOneof(sc.BIJECTOR_TYPE)
  match bijector_type_field:
    case sc.SHIFT_BIJECTOR:
      return backend.bijectors.Shift(
          shift=_deserialize_sequence(bijector_proto.shift.shifts)
      )
    case sc.SCALE_BIJECTOR:
      return backend.bijectors.Scale(
          scale=_deserialize_sequence(bijector_proto.scale.scales),
          log_scale=_deserialize_sequence(bijector_proto.scale.log_scales),
      )
    case sc.RECIPROCAL_BIJECTOR:
      return backend.bijectors.Reciprocal()
    case _:
      raise ValueError(
          f"Unsupported Bijector proto type: {bijector_type_field};"
          f" Bijector proto:\n{bijector_proto}"
      )


def _from_legacy_distribution_proto(
    dist_proto: meridian_pb.Distribution,
) -> backend.tfd.Distribution:
  """Converts a `Distribution` proto to a `Distribution` object."""
  dist_type_field = dist_proto.WhichOneof(sc.DISTRIBUTION_TYPE)
  match dist_type_field:
    case sc.BATCH_BROADCAST_DISTRIBUTION:
      return backend.tfd.BatchBroadcast(
          name=dist_proto.name,
          distribution=_from_legacy_distribution_proto(
              dist_proto.batch_broadcast.distribution
          ),
          with_shape=_from_shape_proto(dist_proto.batch_broadcast.batch_shape),
      )
    case sc.TRANSFORMED_DISTRIBUTION:
      return backend.tfd.TransformedDistribution(
          name=dist_proto.name,
          distribution=_from_legacy_distribution_proto(
              dist_proto.transformed.distribution
          ),
          bijector=_from_legacy_bijector_proto(dist_proto.transformed.bijector),
      )
    case sc.DETERMINISTIC_DISTRIBUTION:
      return backend.tfd.Deterministic(
          name=dist_proto.name,
          loc=_deserialize_sequence(dist_proto.deterministic.locs),
      )
    case sc.HALF_NORMAL_DISTRIBUTION:
      return backend.tfd.HalfNormal(
          name=dist_proto.name,
          scale=_deserialize_sequence(dist_proto.half_normal.scales),
      )
    case sc.LOG_NORMAL_DISTRIBUTION:
      return backend.tfd.LogNormal(
          name=dist_proto.name,
          loc=_deserialize_sequence(dist_proto.log_normal.locs),
          scale=_deserialize_sequence(dist_proto.log_normal.scales),
      )
    case sc.NORMAL_DISTRIBUTION:
      return backend.tfd.Normal(
          name=dist_proto.name,
          loc=_deserialize_sequence(dist_proto.normal.locs),
          scale=_deserialize_sequence(dist_proto.normal.scales),
      )
    case sc.TRUNCATED_NORMAL_DISTRIBUTION:
      if (
          hasattr(dist_proto.truncated_normal, "lows")
          and dist_proto.truncated_normal.lows
      ):
        if dist_proto.truncated_normal.low:
          _show_warning("low", "TruncatedNormal")
        low = _deserialize_sequence(dist_proto.truncated_normal.lows)
      else:
        low = dist_proto.truncated_normal.low

      if (
          hasattr(dist_proto.truncated_normal, "highs")
          and dist_proto.truncated_normal.highs
      ):
        if dist_proto.truncated_normal.high:
          _show_warning("high", "TruncatedNormal")
        high = _deserialize_sequence(dist_proto.truncated_normal.highs)
      else:
        high = dist_proto.truncated_normal.high
      return backend.tfd.TruncatedNormal(
          name=dist_proto.name,
          loc=_deserialize_sequence(dist_proto.truncated_normal.locs),
          scale=_deserialize_sequence(dist_proto.truncated_normal.scales),
          low=low,
          high=high,
      )
    case sc.UNIFORM_DISTRIBUTION:
      if hasattr(dist_proto.uniform, "lows") and dist_proto.uniform.lows:
        if dist_proto.uniform.low:
          _show_warning("low", "Uniform")
        low = _deserialize_sequence(dist_proto.uniform.lows)
      else:
        low = dist_proto.uniform.low

      if hasattr(dist_proto.uniform, "highs") and dist_proto.uniform.highs:
        if dist_proto.uniform.high:
          _show_warning("high", "Uniform")
        high = _deserialize_sequence(dist_proto.uniform.highs)
      else:
        high = dist_proto.uniform.high

      return backend.tfd.Uniform(
          name=dist_proto.name,
          low=low,
          high=high,
      )
    case sc.BETA_DISTRIBUTION:
      return backend.tfd.Beta(
          name=dist_proto.name,
          concentration1=_deserialize_sequence(dist_proto.beta.alpha),
          concentration0=_deserialize_sequence(dist_proto.beta.beta),
      )
    case _:
      raise ValueError(
          f"Unsupported Distribution proto type: {dist_type_field};"
          f" Distribution proto:\n{dist_proto}"
      )


def _show_warning(field_name: str, dist_name: str) -> None:
  warnings.warn(
      f"Both `{field_name}s` and `{field_name}` are specified in"
      f" {dist_name} distribution proto. Prioritizing `{field_name}s` since"
      f" `{field_name}` is deprecated.",
      DeprecationWarning,
  )


def _from_shape_proto(
    shape_proto: tensor_shape_pb2.TensorShapeProto,
) -> backend.TensorShapeInstance:
  """Converts a `TensorShapeProto` to a `TensorShape`."""
  return backend.TensorShape([dim.size for dim in shape_proto.dim])


T = TypeVar("T")


def _deserialize_sequence(args: Sequence[T]) -> T | Sequence[T] | None:
  if not args:
    return None
  return args[0] if len(args) == 1 else list(args)

# copybara: strip_end



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An implementation of `InputDataBuilder` with n-dimensional array primitives."""

import logging
import warnings
from meridian import constants
from meridian.data import input_data_builder
import numpy as np
import xarray as xr


__all__ = [
    'NDArrayInputDataBuilder',
]


class NDArrayInputDataBuilder(input_data_builder.InputDataBuilder):
  """Builds `InputData` from n-dimensional arrays."""

  # Unlike `DataFrameInputDataBuilder`, each piecemeal data has no coordinate
  # information; they're purely data values. It's up to the user to provide
  # coordinates with setter methods from the abstract base class above.
  # Validation is done on each piece w.r.t. dimensional consistency by
  # shape alone.

  def with_kpi(self, nd: np.ndarray) -> 'NDArrayInputDataBuilder':
    """Reads KPI data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos, n_time)`
    -  `(n_time,)` or `(1, n_time)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the KPI data from.

    Returns:
      The `NDArrayInputDataBuilder` with the added KPI data.
    """
    ### Validate ###
    self._validate_coords()
    self._validate_shape(nd)

    ### Transform ###
    self.kpi = xr.DataArray(
        nd,
        dims=[constants.GEO, constants.TIME],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
        },
        name=constants.KPI,
    )
    return self

  def with_controls(
      self, nd: np.ndarray, control_names: list[str]
  ) -> 'NDArrayInputDataBuilder':
    """Reads controls data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos, n_time, n_controls)`
    -  `(n_time, n_controls)` or `(1, n_time, n_controls)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the controls data from.
      control_names: The names of the control variables.

    Returns:
      The `NDArrayInputDataBuilder` with the added controls data.
    """
    ### Validate ###
    self._validate_coords()
    self._validate_shape(nd, control_names)

    ### Transform ###
    self.controls = xr.DataArray(
        nd,
        dims=[constants.GEO, constants.TIME, constants.CONTROL_VARIABLE],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
            constants.CONTROL_VARIABLE: control_names,
        },
        name=constants.CONTROLS,
    )
    return self

  def with_population(self, nd: np.ndarray) -> 'NDArrayInputDataBuilder':
    """Reads population data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos,)`

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the population data from.

    Returns:
      The `NDArrayInputDataBuilder` with the added population data.
    """
    ### Validate ###
    self._validate_coords(is_population=True)
    self._validate_shape(nd, is_population=True)
    ### Transform ###
    self.population = xr.DataArray(
        nd,
        dims=[constants.GEO],
        coords={constants.GEO: self.geos},
        name=constants.POPULATION,
    )

    return self

  def with_revenue_per_kpi(self, nd: np.ndarray) -> 'NDArrayInputDataBuilder':
    """Reads Revenue per KPI data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos, n_time)`
    -  `(n_time,)` or `(1, n_time)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the Reevenue per KPI data from.

    Returns:
      The `NDArrayInputDataBuilder` with the added Revenue per KPI data.
    """
    ### Validate ###
    self._validate_coords()
    self._validate_shape(nd)
    revenue_per_kpi_nd = self._check_revenue_per_kpi_defaults(nd)

    ### Transform ###
    self.revenue_per_kpi = xr.DataArray(
        revenue_per_kpi_nd,
        dims=[constants.GEO, constants.TIME],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
        },
        name=constants.REVENUE_PER_KPI,
    )
    return self

  def with_media(
      self, m_nd: np.ndarray, ms_nd: np.ndarray, media_channels: list[str]
  ) -> 'NDArrayInputDataBuilder':
    """Reads media and media spend data from the ndarrays.

    `m_nd` must be given with the shape:
    -  `(n_geos, n_media_times, n_media_channels)`
    -  `(n_media_times, n_media_channels)` or `(1, n_media_times,
       n_media_channels)` for national model.

    `ms_nd` must be given with the shape:
    -  `(n_geos, n_times, n_media_channels)`
    -  `(n_times, n_media_channels)` or `(1, n_times,
       n_media_channels)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      m_nd: The ndarray that contains dimensional media data.
      ms_nd: The ndarray that contains dimensional media spend data.
      media_channels: The names of the media channels.

    Returns:
      The `NDArrayInputDataBuilder` with the added media and media spend data.
    """
    ### Validate ###
    self._validate_coords(is_media_time=True)
    self._validate_coords(is_media_time=False)
    self._validate_shape(nd=m_nd, dims=media_channels, is_media_time=True)
    self._validate_shape(nd=ms_nd, dims=media_channels, is_media_time=False)

    ### Transform ###
    self.media = xr.DataArray(
        m_nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.MEDIA_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.MEDIA_CHANNEL: media_channels,
        },
        name=constants.MEDIA,
    )
    self.media_spend = xr.DataArray(
        ms_nd,
        dims=[
            constants.GEO,
            constants.TIME,
            constants.MEDIA_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
            constants.MEDIA_CHANNEL: media_channels,
        },
        name=constants.MEDIA_SPEND,
    )
    return self

  def with_reach(
      self,
      r_nd: np.ndarray,
      f_nd: np.ndarray,
      rfs_nd: np.ndarray,
      rf_channels: list[str],
  ) -> 'NDArrayInputDataBuilder':
    """Reads reach, frequency, and rf_spend data from the ndarrays.

    `r_nd` and `f_nd` must be given with the shape:
    -  `(n_geos, n_media_times, n_rf_channels)`
    -  `(n_media_times, n_rf_channels)` or `(1, n_media_times,
       n_rf_channels)` for national model.

    `rfs_nd` must be given with the shape:
    -  `(n_geos, n_times, n_rf_channels)`
    -  `(n_times, n_rf_channels)` or `(1, n_times,
       n_rf_channels)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      r_nd: The ndarray that contains dimensional reach data.
      f_nd: The ndarray that contains dimensional frequency data.
      rfs_nd: The ndarray that contains dimensional rf_spend data.
      rf_channels: The names of the rf channels.

    Returns:
      The `NDArrayInputDataBuilder` with the added reach, frequency, and
      rf_spend data.
    """
    ### Validate ###
    self._validate_coords(is_media_time=True)
    self._validate_coords(is_media_time=False)
    self._validate_shape(nd=r_nd, dims=rf_channels, is_media_time=True)
    self._validate_shape(nd=f_nd, dims=rf_channels, is_media_time=True)
    self._validate_shape(nd=rfs_nd, dims=rf_channels, is_media_time=False)

    ### Transform ###
    self.reach = xr.DataArray(
        r_nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.RF_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.RF_CHANNEL: rf_channels,
        },
        name=constants.REACH,
    )
    self.frequency = xr.DataArray(
        f_nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.RF_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.RF_CHANNEL: rf_channels,
        },
        name=constants.FREQUENCY,
    )
    self.rf_spend = xr.DataArray(
        rfs_nd,
        dims=[
            constants.GEO,
            constants.TIME,
            constants.RF_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
            constants.RF_CHANNEL: rf_channels,
        },
        name=constants.RF_SPEND,
    )
    return self

  def with_organic_media(
      self, nd: np.ndarray, organic_media_channels: list[str]
  ) -> 'NDArrayInputDataBuilder':
    """Reads organic media data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos, n_media_times, n_organic_media_channels)`
    -  `(n_media_times, n_organic_media_channels)` or `(1, n_media_times,
       n_organic_media_channels)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the organic media data from.
      organic_media_channels: The names of the organic media channels.

    Returns:
      The `NDArrayInputDataBuilder` with the added organic media data.
    """
    ### Validate ###
    self._validate_coords(is_media_time=True)
    self._validate_shape(nd=nd, dims=organic_media_channels, is_media_time=True)

    ### Transform ###
    self.organic_media = xr.DataArray(
        nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.ORGANIC_MEDIA_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.ORGANIC_MEDIA_CHANNEL: organic_media_channels,
        },
        name=constants.ORGANIC_MEDIA,
    )
    return self

  def with_organic_reach(
      self, or_nd: np.ndarray, of_nd: np.ndarray, organic_rf_channels: list[str]
  ) -> 'NDArrayInputDataBuilder':
    """Reads organic reach and organic frequency data from the ndarrays.

    `or_nd` and `of_nd` must be given with the shape:
    -  `(n_geos, n_media_times, n_organic_rf_channels)`
    -  `(n_media_times, n_organic_rf_channels)` or `(1, n_media_times,
       n_organic_rf_channels)` for national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      or_nd: The ndarray that contains dimensional reach data.
      of_nd: The ndarray that contains dimensional frequency data.
      organic_rf_channels: The names of the organic rf channels.

    Returns:
      The `NDArrayInputDataBuilder` with the added organic reach and organic
      frequency data.
    """
    ### Validate ###
    self._validate_coords(is_media_time=True)
    self._validate_shape(nd=or_nd, dims=organic_rf_channels, is_media_time=True)
    self._validate_shape(nd=of_nd, dims=organic_rf_channels, is_media_time=True)

    ### Transform ###
    self.organic_reach = xr.DataArray(
        or_nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.ORGANIC_RF_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.ORGANIC_RF_CHANNEL: organic_rf_channels,
        },
        name=constants.ORGANIC_REACH,
    )
    self.organic_frequency = xr.DataArray(
        of_nd,
        dims=[
            constants.GEO,
            constants.MEDIA_TIME,
            constants.ORGANIC_RF_CHANNEL,
        ],
        coords={
            constants.GEO: self.geos,
            constants.MEDIA_TIME: self.media_time_coords,
            constants.ORGANIC_RF_CHANNEL: organic_rf_channels,
        },
        name=constants.ORGANIC_REACH,
    )
    return self

  def with_non_media_treatments(
      self, nd: np.ndarray, non_media_channel_names: list[str]
  ) -> 'NDArrayInputDataBuilder':
    """Reads non-media treatments data from a ndarray.

    `nd` must be given with the shape:
    -  `(n_geos, n_time, n_media_channels)`
    -  `(n_time, n_media_channels)` or `(1, n_time, n_media_channels)` for
    national model.

    If called without a call to .geos() first, the data will be
    assumed to be national-level.

    Args:
      nd: The ndarray to read the non-media treatments data from.
      non_media_channel_names: The names of the non-media channels.

    Returns:
      The `NDArrayInputDataBuilder` with the added non-media treatments data.
    """
    ### Validate ###
    self._validate_coords()
    self._validate_shape(nd, non_media_channel_names)

    ### Transform ###
    self.non_media_treatments = xr.DataArray(
        nd,
        dims=[constants.GEO, constants.TIME, constants.NON_MEDIA_CHANNEL],
        coords={
            constants.GEO: self.geos,
            constants.TIME: self.time_coords,
            constants.NON_MEDIA_CHANNEL: non_media_channel_names,
        },
        name=constants.NON_MEDIA_TREATMENTS,
    )
    return self

  def _validate_coords(
      self, is_population: bool = False, is_media_time: bool = False
  ):
    """Validates that the data has the expected coordinates."""
    if not is_population:
      if is_media_time and self._media_time_coords is None:
        raise ValueError(
            'Media times are required first. Set using .media_time_coords()'
        )
      if not is_media_time and self.time_coords is None:
        raise ValueError(
            'Time coordinates are required first. Set using .time_coords()'
        )
    if self.geos is None:
      logging.warning(
          'No geo coordinates set. Assuming NATIONAL model and geos will be set'
          ' to the default value.'
      )
      self.geos = [constants.NATIONAL_MODEL_DEFAULT_GEO_NAME]

  def _validate_shape(
      self,
      nd: np.ndarray,
      dims: list[str] | None = None,
      is_population: bool = False,
      is_media_time: bool = False,
  ):
    """Validates that the data has the expected shape."""
    # Since all data has a geo dimension (even for national data),
    # Expect the first axis to have the shape of the geo dimension.
    expected_shape = (len(self.geos),)
    detailed_info = f'Expected: {len(self.geos)} geos'
    if not is_population:
      if is_media_time:
        expected_shape += (len(self.media_time_coords),)
        detailed_info += f' x {len(self.media_time_coords)} media times'
      else:
        expected_shape += (len(self.time_coords),)
        detailed_info += f' x {len(self.time_coords)} times'

      if dims is not None:
        if len(dims) != len(set(dims)):
          raise ValueError('given dimensions must be unique.')
        expected_shape += (len(dims),)
        detailed_info += f' x {len(dims)} dims'

    if expected_shape != nd.shape:
      raise ValueError(f'{detailed_info}. Got: {nd.shape}.')

  def _check_revenue_per_kpi_defaults(self, nd: np.ndarray):
    """Sets revenue_per_kpi to default if kpi type is revenue and with_revenue_per_kpi is called."""
    if self._kpi_type == constants.REVENUE:
      warnings.warn(
          'with_revenue_per_kpi was called but kpi_type was set to revenue.'
          ' Assuming revenue per kpi with values [1].'
      )
      return np.ones(nd.shape)
    else:
      return nd



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Marketing analyses converters.

This module defines various classes that convert `MarketingAnalysis`s into flat
dataframes.
"""

import abc
from collections.abc import Iterator, Sequence
import datetime
import functools
import math
import warnings

from meridian import constants as c
from mmm.v1.fit import model_fit_pb2 as fit_pb
from scenarioplanner.converters import mmm
from scenarioplanner.converters.dataframe import constants as dc
from scenarioplanner.converters.dataframe import converter
import pandas as pd


__all__ = [
    "ModelDiagnosticsConverter",
    "ModelFitConverter",
    "MediaOutcomeConverter",
    "MediaSpendConverter",
    "MediaRoiConverter",
]


class ModelDiagnosticsConverter(converter.Converter):
  """Outputs a "ModelDiagnostics" table.

  When called, this converter yields a data frame with the columns:

  *   "Dataset"
  *   "R Squared"
  *   "MAPE"
  *   "wMAPE"
  """

  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    if not self._mmm.model_fit_results:
      return

    model_diagnostics_data = []
    for name, result in self._mmm.model_fit_results.items():
      model_diagnostics_data.append((
          name,
          result.performance.r_squared,
          result.performance.mape,
          result.performance.weighted_mape,
      ))
    yield (
        dc.MODEL_DIAGNOSTICS,
        pd.DataFrame(
            model_diagnostics_data,
            columns=[
                dc.MODEL_DIAGNOSTICS_DATASET_COLUMN,
                dc.MODEL_DIAGNOSTICS_R_SQUARED_COLUMN,
                dc.MODEL_DIAGNOSTICS_MAPE_COLUMN,
                dc.MODEL_DIAGNOSTICS_WMAPE_COLUMN,
            ],
        ),
    )


class ModelFitConverter(converter.Converter):
  """Outputs a "ModelFit" table from an "All Data" (*) `Result` dataset.

  Note: If there is no such result dataset, the first one available is used,
  instead.

  When called, this converter yields a data frame with the columns:

  *   "Time"
      A string formatted with Meridian date format: YYYY-mm-dd
  *   "Expected CI Low"
  *   "Expected CI High"
  *   "Expected"
  *   "Baseline"
  *   "Actual"
  """

  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    if not self._mmm.model_fit_results:
      return

    model_fit_data = []
    for prediction in self._select_model_fit_result().predictions:
      time = datetime.datetime(
          year=prediction.date_interval.start_date.year,
          month=prediction.date_interval.start_date.month,
          day=prediction.date_interval.start_date.day,
      ).strftime(c.DATE_FORMAT)

      if not prediction.predicted_outcome.uncertainties:
        (expected_ci_lo, expected_ci_hi) = (math.nan, math.nan)
      else:
        if len(prediction.predicted_outcome.uncertainties) > 1:
          warnings.warn(
              "More than one `Estimate.uncertainties` found in a"
              " `Prediction.predicted_outcome` in `ModelFit`; processing only"
              " the first confidence interval value."
          )
        uncertainty = prediction.predicted_outcome.uncertainties[0]
        expected_ci_lo = uncertainty.lowerbound
        expected_ci_hi = uncertainty.upperbound
      expected = prediction.predicted_outcome.value
      actual = prediction.actual_value

      baseline = prediction.predicted_baseline.value

      model_fit_data.append(
          (time, expected_ci_lo, expected_ci_hi, expected, baseline, actual)
      )

    yield (
        dc.MODEL_FIT,
        pd.DataFrame(
            model_fit_data,
            columns=[
                dc.MODEL_FIT_TIME_COLUMN,
                dc.MODEL_FIT_EXPECTED_CI_LOW_COLUMN,
                dc.MODEL_FIT_EXPECTED_CI_HIGH_COLUMN,
                dc.MODEL_FIT_EXPECTED_COLUMN,
                dc.MODEL_FIT_BASELINE_COLUMN,
                dc.MODEL_FIT_ACTUAL_COLUMN,
            ],
        ),
    )

  def _select_model_fit_result(self) -> fit_pb.Result:
    """Returns the model fit `Result` dataset with name "All Data".

    Or else, first available.
    """
    model_fit_results = self._mmm.model_fit_results
    if not model_fit_results:
      raise ValueError("Must have at least one `ModelFit.results` value.")
    if c.ALL_DATA in model_fit_results:
      result = model_fit_results[c.ALL_DATA]
    else:
      result = self._mmm.model_fit.results[0]
      warnings.warn(f"Using a model fit `Result` with name: '{result.name}'")

    return result


class _MarketingAnalysisConverter(converter.Converter, abc.ABC):
  """An abstract class for dealing with `MarketingAnalysis`."""

  @functools.cached_property
  def _is_revenue_kpi(self) -> bool:
    """Returns true if analyses are using revenue KPI.

    This is done heuristically: by looking at the (presumed existing) "baseline"
    `NonMediaAnalysis` proto and seeing if `revenue_kpi` field is defined. If it
    is, we assume that all other media analyses must have their `revenue_kpi`
    fields defined, too.

    Likewise: if the baseline analysis defines `non_revenue_kpi` and it does not
    define `revenue_kpi`, we assume that all other media analyses are based on
    a non-revenue KPI.

    Note: This means that this output converter can only work with one type of
    KPI as a whole. If a channel's media analysis has both revenue- and
    nonrevenue-type KPI defined, for example, only the former will be outputted.
    """
    baseline_analysis = self._mmm.tagged_marketing_analyses[
        dc.ANALYSIS_TAG_ALL
    ].baseline_analysis
    return baseline_analysis.maybe_revenue_outcome is not None

  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    if not self._mmm.marketing_analyses:
      return

    yield from self._handle_marketing_analyses(self._mmm.marketing_analyses)

  def _handle_marketing_analyses(
      self, analyses: Sequence[mmm.MarketingAnalysis]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    raise NotImplementedError()


class MediaOutcomeConverter(_MarketingAnalysisConverter):
  """Outputs a "MediaOutcome" table.

  When called, this converter yields a data frame with the columns:

  *   "Channel Index"
      This is to ensure "baseline" and "All Channels" can be sorted to appear
      first and last, respectively, in LS dashboard.
  *   "Channel"
  *   "Incremental Outcome"
  *   "Contribution Share"
  *   "Analysis Period"
      A human-readable analysis period.
  *   "Analysis Date Start"
      A string formatted with Meridian date format: YYYY-mm-dd
  *   "Analysis Date End"
      A string formatted with Meridian date format: YYYY-mm-dd

  Note: If the underlying model analysis works with a revenue-type KPI (i.e.
  dollar value), then all values in the columns of the output table should be
  interpreted the same. Likewise, for non-revenue type KPI. While some
  channels may define their outcome analyses in terms of both revenue- and
  nonrevenue-type semantics, the output table here remains uniform.
  """

  def _handle_marketing_analyses(
      self, analyses: Sequence[mmm.MarketingAnalysis]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    media_outcome_data = []
    for marketing_analysis in analyses:
      date_start, date_end = marketing_analysis.analysis_date_interval_str

      baseline_outcome: mmm.Outcome = (
          marketing_analysis.baseline_analysis.revenue_outcome
          if self._is_revenue_kpi
          else marketing_analysis.baseline_analysis.non_revenue_outcome
      )
      # "contribution" == incremental outcome
      baseline_contrib = baseline_outcome.contribution_pb.value.value
      baseline_contrib_share = baseline_outcome.contribution_pb.share.value

      media_outcome_data.append((
          dc.MEDIA_OUTCOME_BASELINE_PSEUDO_CHANNEL_INDEX,
          c.BASELINE,
          baseline_contrib,
          baseline_contrib_share,
          marketing_analysis.tag,
          date_start,
          date_end,
      ))
      media_analyses = list(
          marketing_analysis.channel_mapped_media_analyses.items()
      )
      non_media_analyses = list(filter(
          lambda x: x[0] != c.BASELINE,
          list(marketing_analysis.channel_mapped_non_media_analyses.items()),
      ))
      all_analyses = media_analyses + non_media_analyses
      for channel, media_analysis in all_analyses:
        channel_index = (
            dc.MEDIA_OUTCOME_ALL_CHANNELS_PSEUDO_CHANNEL_INDEX
            if channel == c.ALL_CHANNELS
            else dc.MEDIA_OUTCOME_CHANNEL_INDEX
        )
        # Note: use the same revenue- or nonrevenue-type outcome analysis as the
        # baseline's.
        try:
          channel_outcome: mmm.Outcome = (
              media_analysis.revenue_outcome
              if self._is_revenue_kpi
              else media_analysis.non_revenue_outcome
          )
        except ValueError:
          warnings.warn(
              f"No {'' if self._is_revenue_kpi else 'non'}revenue-type"
              " `Outcome` found in the channel media analysis for"
              f' "{channel}"'
          )
          channel_contrib = math.nan
          channel_contrib_share = math.nan
        else:
          channel_contrib = channel_outcome.contribution_pb.value.value
          channel_contrib_share = channel_outcome.contribution_pb.share.value

        media_outcome_data.append((
            channel_index,
            channel,
            channel_contrib,
            channel_contrib_share,
            marketing_analysis.tag,
            date_start,
            date_end,
        ))

    yield (
        dc.MEDIA_OUTCOME,
        pd.DataFrame(
            media_outcome_data,
            columns=[
                dc.MEDIA_OUTCOME_CHANNEL_INDEX_COLUMN,
                dc.MEDIA_OUTCOME_CHANNEL_COLUMN,
                dc.MEDIA_OUTCOME_INCREMENTAL_OUTCOME_COLUMN,
                dc.MEDIA_OUTCOME_CONTRIBUTION_SHARE_COLUMN,
                dc.ANALYSIS_PERIOD_COLUMN,  # using the `tag` field
                dc.ANALYSIS_DATE_START_COLUMN,
                dc.ANALYSIS_DATE_END_COLUMN,
            ],
        ),
    )


class MediaSpendConverter(_MarketingAnalysisConverter):
  """Outputs a "MediaSpend" table.

  When called, this converter yields a data frame with the columns:

  *   "Channel"
  *   "Value"
  *   "Label"
      A human-readable label on what "Value" represents
  *   "Analysis Period"
      A human-readable analysis period.
  *   "Analysis Date Start"
      A string formatted with Meridian date format: YYYY-mm-dd
  *   "Analysis Date End"
      A string formatted with Meridian date format: YYYY-mm-dd

  Note: If the underlying model analysis works with a revenue-type KPI (i.e.
  dollar value), then all values in the columns of the output table should
  be interpreted the same. Likewise, for non-revenue type KPI. While some
  channels may define their outcome analyses in terms of both revenue- and
  nonrevenue-type semantics, the output table here remains uniform.
  """

  _share_value_column_index = 1
  _label_column_index = 2

  def _handle_marketing_analyses(
      self, analyses: Sequence[mmm.MarketingAnalysis]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    media_spend_data = []

    for analysis in analyses:
      date_start, date_end = analysis.analysis_date_interval_str
      data = []
      outcome_share_norm_term = 0.0

      for (
          channel,
          media_analysis,
      ) in analysis.channel_mapped_media_analyses.items():
        # Ignore the "All Channels" pseudo-channel.
        if channel == c.ALL_CHANNELS:
          continue

        spend_share = media_analysis.spend_info_pb.spend_share

        try:
          channel_outcome: mmm.Outcome = (
              media_analysis.revenue_outcome
              if self._is_revenue_kpi
              else media_analysis.non_revenue_outcome
          )
        except ValueError:
          warnings.warn(
              f"No {'' if self._is_revenue_kpi else 'non'}revenue-type"
              " `Outcome` found in the channel media analysis for"
              f' "{channel}"'
          )
          outcome_share = math.nan
        else:
          outcome_share = channel_outcome.contribution_pb.share.value
          outcome_share_norm_term += outcome_share

        data.append([
            channel,
            spend_share,
            dc.MEDIA_SPEND_LABEL_SPEND_SHARE,
            analysis.tag,
            date_start,
            date_end,
        ])
        data.append([
            channel,
            outcome_share,
            (
                dc.MEDIA_SPEND_LABEL_REVENUE_SHARE
                if self._is_revenue_kpi
                else dc.MEDIA_SPEND_LABEL_KPI_SHARE
            ),
            analysis.tag,
            date_start,
            date_end,
        ])

      # Looker Studio media spend/revenue share charts expect the "revenue
      # share" values to be normalized to 100%. This normaliztion provides
      # additional information to what the contribution waterfall chart already
      # provides.
      for d in data:
        if d[self._label_column_index] == dc.MEDIA_SPEND_LABEL_SPEND_SHARE:
          continue
        d[self._share_value_column_index] /= outcome_share_norm_term

      media_spend_data.extend(data)

    yield (
        dc.MEDIA_SPEND,
        pd.DataFrame(
            media_spend_data,
            columns=[
                dc.MEDIA_SPEND_CHANNEL_COLUMN,
                dc.MEDIA_SPEND_SHARE_VALUE_COLUMN,
                dc.MEDIA_SPEND_LABEL_COLUMN,
                dc.ANALYSIS_PERIOD_COLUMN,  # using the `tag` field
                dc.ANALYSIS_DATE_START_COLUMN,
                dc.ANALYSIS_DATE_END_COLUMN,
            ],
        ),
    )


class MediaRoiConverter(_MarketingAnalysisConverter):
  """Outputs a "MediaROI" table.

  When called, this converter yields a data frame with the columns:

    *   "Channel"
    *   "Spend"
    *   "Effectiveness"
    *   "ROI"
    *   "ROI CI Low"
        The confidence interval (low) of "ROI" above.
    *   "ROI CI High"
        The confidence interval (high) of "ROI" above.
    *   "Marginal ROI"
    *   "Is Revenue KPI"
        A boolean indicating whether "ROI" refers to revenue or generic KPI.
    *   "Analysis Period"
        A human-readable analysis period.
    *   "Analysis Date Start"
        A string formatted with Meridian date format: YYYY-mm-dd
    *   "Analysis Date End"
        A string formatted with Meridian date format: YYYY-mm-dd

  Note: If the underlying model analysis works with a revenue-type KPI (i.e.
  dollar value), then all values in the columns of the output table should
  be interpreted the same. Likewise, for non-revenue type KPI. While some
  channels may define their outcome analyses in terms of both revenue- and
  nonrevenue-type semantics, the output table here remains uniform.
  """

  def _handle_marketing_analyses(
      self, analyses: Sequence[mmm.MarketingAnalysis]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    media_roi_data = []
    for analysis in analyses:
      date_start, date_end = analysis.analysis_date_interval_str

      for (
          channel,
          media_analysis,
      ) in analysis.channel_mapped_media_analyses.items():
        # Ignore the "All Channels" pseudo-channel.
        if channel == c.ALL_CHANNELS:
          continue

        spend = media_analysis.spend_info_pb.spend

        try:
          channel_outcome: mmm.Outcome = (
              media_analysis.revenue_outcome
              if self._is_revenue_kpi
              else media_analysis.non_revenue_outcome
          )
        except ValueError as exc:
          raise ValueError(
              f"No {'' if self._is_revenue_kpi else 'non'}revenue-type"
              " `Outcome` found in the channel media analysis for"
              f' "{channel}"'
          ) from exc
        else:
          effectiveness = channel_outcome.effectiveness_pb.value.value
          roi_estimate = channel_outcome.roi_pb
          if not roi_estimate.uncertainties:
            (roi_ci_lo, roi_ci_hi) = (math.nan, math.nan)
          else:
            if len(roi_estimate.uncertainties) > 1:
              warnings.warn(
                  "More than one `Estimate.uncertainties` found in an"
                  ' `Outcome.revenue_outcome.roi` in channel "{channel}".'
                  " Using the first confidence interval value."
              )
            uncertainty = roi_estimate.uncertainties[0]
            roi_ci_lo = uncertainty.lowerbound
            roi_ci_hi = uncertainty.upperbound
          roi = roi_estimate.value
          marginal_roi = channel_outcome.marginal_roi_pb.value
          is_revenue_kpi = channel_outcome.is_revenue_kpi

        media_roi_data.append([
            channel,
            spend,
            effectiveness,
            roi,
            roi_ci_lo,
            roi_ci_hi,
            marginal_roi,
            is_revenue_kpi,
            analysis.tag,
            date_start,
            date_end,
        ])

    yield (
        dc.MEDIA_ROI,
        pd.DataFrame(
            media_roi_data,
            columns=[
                dc.MEDIA_ROI_CHANNEL_COLUMN,
                dc.MEDIA_ROI_SPEND_COLUMN,
                dc.MEDIA_ROI_EFFECTIVENESS_COLUMN,
                dc.MEDIA_ROI_ROI_COLUMN,
                dc.MEDIA_ROI_ROI_CI_LOW_COLUMN,
                dc.MEDIA_ROI_ROI_CI_HIGH_COLUMN,
                dc.MEDIA_ROI_MARGINAL_ROI_COLUMN,
                dc.MEDIA_ROI_IS_REVENUE_KPI_COLUMN,
                dc.ANALYSIS_PERIOD_COLUMN,
                dc.ANALYSIS_DATE_START_COLUMN,
                dc.ANALYSIS_DATE_END_COLUMN,
            ],
        ),
    )


CONVERTERS = [
    # These converters create tables for the model analysis charts to use:
    ModelDiagnosticsConverter,
    ModelFitConverter,
    MediaOutcomeConverter,
    MediaSpendConverter,
    MediaRoiConverter,
]



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meridian module for the geo-level Bayesian hierarchical media mix model."""

import collections
from collections.abc import Mapping, Sequence
import functools
import os
import warnings

import arviz as az
import joblib
from meridian import backend
from meridian import constants
from meridian.data import input_data as data
from meridian.data import time_coordinates as tc
from meridian.model import adstock_hill
from meridian.model import context
from meridian.model import equations
from meridian.model import knots
from meridian.model import media
from meridian.model import posterior_sampler
from meridian.model import prior_distribution
from meridian.model import prior_sampler
from meridian.model import spec
from meridian.model import transformers
from meridian.model.eda import eda_engine
from meridian.model.eda import eda_outcome
from meridian.model.eda import eda_spec as eda_spec_module
import numpy as np

__all__ = [
    "MCMCSamplingError",
    "MCMCOOMError",
    "Meridian",
    "ModelFittingError",
    "NotFittedModelError",
    "save_mmm",
    "load_mmm",
]


class ModelFittingError(Exception):
  """Model has critical issues preventing fitting."""


class NotFittedModelError(Exception):
  """Model has not been fitted."""


MCMCSamplingError = posterior_sampler.MCMCSamplingError
MCMCOOMError = posterior_sampler.MCMCOOMError


def _warn_setting_national_args(**kwargs):
  """Raises a warning if a geo argument is found in kwargs."""
  for kwarg, value in kwargs.items():
    if (
        kwarg in constants.NATIONAL_MODEL_SPEC_ARGS
        and value is not constants.NATIONAL_MODEL_SPEC_ARGS[kwarg]
    ):
      warnings.warn(
          f"In a nationally aggregated model, the `{kwarg}` will be reset to"
          f" `{constants.NATIONAL_MODEL_SPEC_ARGS[kwarg]}`."
      )


class Meridian:
  """Contains the main functionality for fitting the Meridian MMM model.

  Attributes:
    input_data: An `InputData` object containing the input data for the model.
    model_spec: A `ModelSpec` object containing the model specification.
    model_context: A `ModelContext` object containing the model context.
    equations: A `ModelEquations` object containing stateless mathematical
      functions and utilities for Meridian MMM.
    inference_data: A _mutable_ `arviz.InferenceData` object containing the
      resulting data from fitting the model.
    eda_engine: An `EDAEngine` object containing the EDA engine.
    eda_spec: An `EDASpec` object containing the EDA specification.
    eda_outcomes: A list of `EDAOutcome` objects containing the outcomes from
      running critical EDA checks.
    n_geos: Number of geos in the data.
    n_media_channels: Number of media channels in the data.
    n_rf_channels: Number of reach and frequency (RF) channels in the data.
    n_organic_media_channels: Number of organic media channels in the data.
    n_organic_rf_channels: Number of organic reach and frequency (RF) channels
      in the data.
    n_controls: Number of control variables in the data.
    n_non_media_channels: Number of non-media treatment channels in the data.
    n_times: Number of time periods in the KPI or spend data.
    n_media_times: Number of time periods in the media data.
    is_national: A boolean indicating whether the data is national (single geo)
      or not (multiple geos).
    knot_info: A `KnotInfo` derived from input data and model spec.
    kpi: A tensor constructed from `input_data.kpi`.
    revenue_per_kpi: A tensor constructed from `input_data.revenue_per_kpi`. If
      `input_data.revenue_per_kpi` is None, then this is also None.
    controls: A tensor constructed from `input_data.controls`.
    non_media_treatments: A tensor constructed from
      `input_data.non_media_treatments`.
    population: A tensor constructed from `input_data.population`.
    media_tensors: A collection of media tensors derived from `input_data`.
    rf_tensors: A collection of Reach & Frequency (RF) media tensors.
    organic_media_tensors: A collection of organic media tensors.
    organic_rf_tensors: A collection of organic Reach & Frequency (RF) media
      tensors.
    total_spend: A tensor containing total spend, including
      `media_tensors.media_spend` and `rf_tensors.rf_spend`.
    total_outcome: A tensor containing the total outcome, aggregated over geos
      and times.
    controls_transformer: A `ControlsTransformer` to scale controls tensors
      using the model's controls data.
    non_media_transformer: A `CenteringAndScalingTransformer` to scale non-media
      treatmenttensors using the model's non-media treatment data.
    kpi_transformer: A `KpiTransformer` to scale KPI tensors using the model's
      KPI data.
    controls_scaled: The controls tensor after pre-modeling transformations
      including population scaling (for variables with
      `ModelSpec.control_population_scaling_id` set to `True`), centering by the
      mean, and scaling by the standard deviation.
    non_media_treatments_scaled: The non-media treatment tensor after
      pre-modeling transformations including population scaling (for variables
      with `ModelSpec.non_media_population_scaling_id` set to `True`), centering
      by the mean, and scaling by the standard deviation.
    kpi_scaled: The KPI tensor after pre-modeling transformations including
      population scaling, centering by the mean, and scaling by the standard
      deviation.
    media_effects_dist: A string to specify the distribution of media random
      effects across geos.
    unique_sigma_for_each_geo: A boolean indicating whether to use a unique
      residual variance for each geo.
    prior_broadcast: A `PriorDistribution` object containing broadcasted
      distributions.
    baseline_geo_idx: The index of the baseline geo.
    holdout_id: A tensor containing the holdout id, if present.
  """

  def __init__(
      self,
      input_data: data.InputData,
      model_spec: spec.ModelSpec | None = None,
      inference_data: (
          az.InferenceData | None
      ) = None,  # for deserializer use only
      eda_spec: eda_spec_module.EDASpec = eda_spec_module.EDASpec(),
  ):
    self._inference_data = (
        inference_data if inference_data else az.InferenceData()
    )
    self._model_context = context.ModelContext(
        input_data=input_data,
        model_spec=model_spec if model_spec else spec.ModelSpec(),
    )
    self._equations = equations.ModelEquations(self._model_context)

    self._eda_spec = eda_spec

    self._validate_injected_inference_data()

    if self.is_national:
      _warn_setting_national_args(
          media_effects_dist=self.model_spec.media_effects_dist,
          unique_sigma_for_each_geo=self.model_spec.unique_sigma_for_each_geo,
      )
    self._validate_kpi_variability()

  @property
  def input_data(self) -> data.InputData:
    return self._model_context.input_data

  @property
  def model_spec(self) -> spec.ModelSpec:
    return self._model_context.model_spec

  @property
  def model_context(self) -> context.ModelContext:
    return self._model_context

  @property
  def equations(self) -> equations.ModelEquations:
    return self._equations

  @property
  def inference_data(self) -> az.InferenceData:
    return self._inference_data

  @functools.cached_property
  def eda_engine(self) -> eda_engine.EDAEngine:
    return eda_engine.EDAEngine(self, spec=self._eda_spec)

  @property
  def eda_spec(self) -> eda_spec_module.EDASpec:
    return self._eda_spec

  @property
  def eda_outcomes(self) -> Sequence[eda_outcome.EDAOutcome]:
    return self.eda_engine.run_all_critical_checks()

  @functools.cached_property
  def media_tensors(self) -> media.MediaTensors:
    return self._model_context.media_tensors

  @functools.cached_property
  def rf_tensors(self) -> media.RfTensors:
    return self._model_context.rf_tensors

  @functools.cached_property
  def organic_media_tensors(self) -> media.OrganicMediaTensors:
    return self._model_context.organic_media_tensors

  @functools.cached_property
  def organic_rf_tensors(self) -> media.OrganicRfTensors:
    return self._model_context.organic_rf_tensors

  @functools.cached_property
  def kpi(self) -> backend.Tensor:
    return self._model_context.kpi

  @functools.cached_property
  def revenue_per_kpi(self) -> backend.Tensor | None:
    return self._model_context.revenue_per_kpi

  @functools.cached_property
  def controls(self) -> backend.Tensor | None:
    return self._model_context.controls

  @functools.cached_property
  def non_media_treatments(self) -> backend.Tensor | None:
    return self._model_context.non_media_treatments

  @functools.cached_property
  def population(self) -> backend.Tensor:
    return self._model_context.population

  @functools.cached_property
  def total_spend(self) -> backend.Tensor:
    return self._model_context.total_spend

  @functools.cached_property
  def total_outcome(self) -> backend.Tensor:
    return self._model_context.total_outcome

  @property
  def n_geos(self) -> int:
    return self._model_context.n_geos

  @property
  def n_media_channels(self) -> int:
    return self._model_context.n_media_channels

  @property
  def n_rf_channels(self) -> int:
    return self._model_context.n_rf_channels

  @property
  def n_organic_media_channels(self) -> int:
    return self._model_context.n_organic_media_channels

  @property
  def n_organic_rf_channels(self) -> int:
    return self._model_context.n_organic_rf_channels

  @property
  def n_controls(self) -> int:
    return self._model_context.n_controls

  @property
  def n_non_media_channels(self) -> int:
    return self._model_context.n_non_media_channels

  @property
  def n_times(self) -> int:
    return self._model_context.n_times

  @property
  def n_media_times(self) -> int:
    return self._model_context.n_media_times

  @property
  def is_national(self) -> bool:
    return self._model_context.is_national

  @functools.cached_property
  def knot_info(self) -> knots.KnotInfo:
    return self._model_context.knot_info

  @functools.cached_property
  def controls_transformer(
      self,
  ) -> transformers.CenteringAndScalingTransformer | None:
    return self._model_context.controls_transformer

  @functools.cached_property
  def non_media_transformer(
      self,
  ) -> transformers.CenteringAndScalingTransformer | None:
    return self._model_context.non_media_transformer

  @functools.cached_property
  def kpi_transformer(self) -> transformers.KpiTransformer:
    return self._model_context.kpi_transformer

  @functools.cached_property
  def controls_scaled(self) -> backend.Tensor | None:
    return self._model_context.controls_scaled

  @functools.cached_property
  def non_media_treatments_normalized(self) -> backend.Tensor | None:
    """Normalized non-media treatments.

    The non-media treatments values are scaled by population (for channels where
    `non_media_population_scaling_id` is `True`) and normalized by centering and
    scaling with means and standard deviations.
    """
    return self._model_context.non_media_treatments_normalized

  @functools.cached_property
  def kpi_scaled(self) -> backend.Tensor:
    return self._model_context.kpi_scaled

  @functools.cached_property
  def media_effects_dist(self) -> str:
    return self._model_context.media_effects_dist

  @functools.cached_property
  def unique_sigma_for_each_geo(self) -> bool:
    return self._model_context.unique_sigma_for_each_geo

  @functools.cached_property
  def baseline_geo_idx(self) -> int:
    """Returns the index of the baseline geo."""
    return self._model_context.baseline_geo_idx

  @functools.cached_property
  def holdout_id(self) -> backend.Tensor | None:
    return self._model_context.holdout_id

  @functools.cached_property
  def adstock_decay_spec(self) -> adstock_hill.AdstockDecaySpec:
    """Returns `AdstockDecaySpec` object with correctly mapped channels."""
    return self._model_context.adstock_decay_spec

  @functools.cached_property
  def prior_broadcast(self) -> prior_distribution.PriorDistribution:
    """Returns broadcasted `PriorDistribution` object."""
    return self._model_context.prior_broadcast

  @functools.cached_property
  def prior_sampler_callable(self) -> prior_sampler.PriorDistributionSampler:
    """A `PriorDistributionSampler` callable bound to this model."""
    return prior_sampler.PriorDistributionSampler(self)

  @functools.cached_property
  def posterior_sampler_callable(
      self,
  ) -> posterior_sampler.PosteriorMCMCSampler:
    """A `PosteriorMCMCSampler` callable bound to this model."""
    return posterior_sampler.PosteriorMCMCSampler(self)

  # TODO: Deprecate this method in favor of the one in
  # `equations.py`.
  def compute_non_media_treatments_baseline(
      self,
      non_media_baseline_values: Sequence[str | float] | None = None,
  ) -> backend.Tensor:
    """Computes the baseline for each non-media treatment channel.

    Args:
      non_media_baseline_values: Optional list of shape
        `(n_non_media_channels,)`. Each element is either a float (which means
        that the fixed value will be used as baseline for the given channel) or
        one of the strings "min" or "max" (which mean that the global minimum or
        maximum value will be used as baseline for the values of the given
        non_media treatment channel). If float values are provided, it is
        expected that they are scaled by population for the channels where
        `model_spec.non_media_population_scaling_id` is `True`. If `None`, the
        `model_spec.non_media_baseline_values` is used, which defaults to the
        minimum value for each non_media treatment channel.

    Returns:
      A tensor of shape `(n_non_media_channels,)` containing the
      baseline values for each non-media treatment channel.
    """
    return self.equations.compute_non_media_treatments_baseline(
        non_media_baseline_values=non_media_baseline_values
    )

  def expand_selected_time_dims(
      self,
      start_date: tc.Date = None,
      end_date: tc.Date = None,
  ) -> list[str] | None:
    """Validates and returns time dimension values based on the selected times.

    If both `start_date` and `end_date` are None, returns None. If specified,
    both `start_date` and `end_date` are inclusive, and must be present in the
    time coordinates of the input data.

    Args:
      start_date: Start date of the selected time period. If None, implies the
        earliest time dimension value in the input data.
      end_date: End date of the selected time period. If None, implies the
        latest time dimension value in the input data.

    Returns:
      A list of time dimension values (as Meridian-formatted strings) in the
      input data within the selected time period, or do nothing and pass through
      None if both arguments are Nones, or if `start_date` and `end_date`
      correspond to the entire time range in the input data.

    Raises:
      ValueError if `start_date` or `end_date` is not in the input data time
      dimensions.
    """
    expanded = self.input_data.time_coordinates.expand_selected_time_dims(
        start_date=start_date, end_date=end_date
    )
    if expanded is None:
      return None
    return [date.strftime(constants.DATE_FORMAT) for date in expanded]

  def _validate_injected_inference_data(self):
    """Validates that the injected inference data has correct shapes.

    Raises:
      ValueError: If the injected `InferenceData` has incorrect shapes.
    """
    if hasattr(self.inference_data, constants.PRIOR):
      self._validate_injected_inference_data_group(
          self.inference_data, constants.PRIOR
      )
    if hasattr(self.inference_data, constants.POSTERIOR):
      self._validate_injected_inference_data_group(
          self.inference_data, constants.POSTERIOR
      )

  def _validate_injected_inference_data_group_coord(
      self,
      inference_data: az.InferenceData,
      group: str,
      coord: str,
      expected_size: int,
  ):
    """Validates that the injected inference data group coordinate has the expected size.

    Args:
      inference_data: The injected `InferenceData` to be validated.
      group: The group of the coordinate to be validated.
      coord: The coordinate to be validated.
      expected_size: The expected size of the coordinate.

    Raises:
      ValueError: If the injected `InferenceData` has incorrect size for the
      coordinate.
    """

    injected_size = (
        inference_data[group].coords[coord].size
        if coord in inference_data[group].coords
        else 0
    )
    if injected_size != expected_size:
      raise ValueError(
          f"Injected inference data {group} has incorrect coordinate '{coord}':"
          f" expected {expected_size}, got {injected_size}"
      )

  def _validate_injected_inference_data_group(
      self,
      inference_data: az.InferenceData,
      group: str,
  ):
    """Validates that the injected inference data group has correct shapes.

    Args:
      inference_data: The injected `InferenceData` to be validated.
      group: The group of the coordinate to be validated.

    Raises:
      ValueError: If the injected `InferenceData` has incorrect shapes.
    """

    self._validate_injected_inference_data_group_coord(
        inference_data, group, constants.GEO, self.n_geos
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.CONTROL_VARIABLE,
        self.n_controls,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.KNOTS,
        self.knot_info.n_knots,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data, group, constants.TIME, self.n_times
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.MEDIA_CHANNEL,
        self.n_media_channels,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.RF_CHANNEL,
        self.n_rf_channels,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.ORGANIC_MEDIA_CHANNEL,
        self.n_organic_media_channels,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.ORGANIC_RF_CHANNEL,
        self.n_organic_rf_channels,
    )
    self._validate_injected_inference_data_group_coord(
        inference_data,
        group,
        constants.NON_MEDIA_CHANNEL,
        self.n_non_media_channels,
    )

  def _validate_kpi_variability(self):
    """Validates the KPI variability."""
    if self.eda_engine.kpi_has_variability:
      return
    kpi = self.eda_engine.kpi_scaled_da.name

    if (
        self.n_media_channels > 0
        and self.model_spec.effective_media_prior_type
        in constants.PAID_MEDIA_ROI_PRIOR_TYPES
    ):
      raise ValueError(
          f"`{kpi}` cannot be constant with"
          " `media_prior_type` ="
          f' "{self.model_spec.effective_media_prior_type}".'
      )
    if (
        self.n_rf_channels > 0
        and self.model_spec.effective_rf_prior_type
        in constants.PAID_MEDIA_ROI_PRIOR_TYPES
    ):
      raise ValueError(
          f"`{kpi}` cannot be constant with"
          f' `rf_prior_type` = "{self.model_spec.effective_rf_prior_type}".'
      )
    if (
        self.n_organic_media_channels > 0
        and self.model_spec.organic_media_prior_type
        in [constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION]
    ):
      raise ValueError(
          f"`{kpi}` cannot be constant with"
          " `organic_media_prior_type` ="
          f' "{self.model_spec.organic_media_prior_type}".'
      )
    if (
        self.n_organic_rf_channels > 0
        and self.model_spec.organic_rf_prior_type
        in [constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION]
    ):
      raise ValueError(
          f"`{kpi}` cannot be constant with"
          " `organic_rf_prior_type` ="
          f' "{self.model_spec.organic_rf_prior_type}".'
      )
    if (
        self.n_non_media_channels > 0
        and self.model_spec.non_media_treatments_prior_type
        in [constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION]
    ):
      raise ValueError(
          f"`{kpi}` cannot be constant with"
          " `non_media_treatments_prior_type` ="
          f' "{self.model_spec.non_media_treatments_prior_type}".'
      )

  # TODO: Deprecate in favor of ModelEquations.adstock_hill_rf.
  def linear_predictor_counterfactual_difference_media(
      self,
      media_transformed: backend.Tensor,
      alpha_m: backend.Tensor,
      ec_m: backend.Tensor,
      slope_m: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates linear predictor counterfactual difference for non-RF media.

    For non-RF media variables (paid or organic), this function calculates the
    linear predictor difference between the treatment variable and its
    counterfactual. "Linear predictor" refers to the output of the hill/adstock
    function, which is multiplied by the geo-level coefficient.

    This function does the calculation efficiently by only calculating calling
    the hill/adstock function if the prior counterfactual is not all zeros.

    Args:
      media_transformed: The output of the hill/adstock function for actual
        historical media data.
      alpha_m: The adstock alpha parameter values.
      ec_m: The adstock ec parameter values.
      slope_m: The adstock hill slope parameter values.

    Returns:
      The linear predictor difference between the treatment variable and its
      counterfactual.
    """
    return self.equations.linear_predictor_counterfactual_difference_media(
        media_transformed=media_transformed,
        alpha_m=alpha_m,
        ec_m=ec_m,
        slope_m=slope_m,
    )

  # TODO: Deprecate in favor of
  # ModelEquations.linear_predictor_counterfactual_difference_rf.
  def linear_predictor_counterfactual_difference_rf(
      self,
      rf_transformed: backend.Tensor,
      alpha_rf: backend.Tensor,
      ec_rf: backend.Tensor,
      slope_rf: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates linear predictor counterfactual difference for RF media.

    For RF media variables (paid or organic), this function calculates the
    linear predictor difference between the treatment variable and its
    counterfactual. "Linear predictor" refers to the output of the hill/adstock
    function, which is multiplied by the geo-level coefficient.

    This function does the calculation efficiently by only calculating calling
    the hill/adstock function if the prior counterfactual is not all zeros.

    Args:
      rf_transformed: The output of the hill/adstock function for actual
        historical media data.
      alpha_rf: The adstock alpha parameter values.
      ec_rf: The adstock ec parameter values.
      slope_rf: The adstock hill slope parameter values.

    Returns:
      The linear predictor difference between the treatment variable and its
      counterfactual.
    """
    return self.equations.linear_predictor_counterfactual_difference_rf(
        rf_transformed=rf_transformed,
        alpha_rf=alpha_rf,
        ec_rf=ec_rf,
        slope_rf=slope_rf,
    )

  # TODO: Deprecate in favor of ModelEquations.calculate_beta_x.
  def calculate_beta_x(
      self,
      is_non_media: bool,
      incremental_outcome_x: backend.Tensor,
      linear_predictor_counterfactual_difference: backend.Tensor,
      eta_x: backend.Tensor,
      beta_gx_dev: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates coefficient mean parameter for any treatment variable type.

    The "beta_x" in the function name refers to the coefficient mean parameter
    of any treatment variable. The "x" can represent "m", "rf", "om", or "orf".
    This function can also be used to calculate "gamma_n" for any non-media
    treatments.

    Args:
      is_non_media: Boolean indicating whether the treatment variable is a
        non-media treatment. This argument is used to determine whether the
        coefficient random effects are normal or log-normal. If `True`, then
        random effects are assumed to be normal. Otherwise, the distribution is
        inferred from `self.media_effects_dist`.
      incremental_outcome_x: The incremental outcome of the treatment variable,
        which depends on the parameter values of a particular prior or posterior
        draw. The "_x" indicates that this is a tensor with length equal to the
        dimension of the treatment variable.
      linear_predictor_counterfactual_difference: The difference between the
        treatment variable and its counterfactual on the linear predictor scale.
        "Linear predictor" refers to the quantity that is multiplied by the
        geo-level coefficient. For media variables, this is the output of the
        hill/adstock transformation function. For non-media treatments, this is
        simply the treatment variable after centering/scaling transformations.
        This tensor has dimensions for geo, time, and channel.
      eta_x: The random effect standard deviation parameter values. For media
        variables, the "x" represents "m", "rf", "om", or "orf". For non-media
        treatments, this argument should be set to `xi_n`, which is analogous to
        "eta".
      beta_gx_dev: The latent standard normal parameter values of the geo-level
        coefficients. For media variables, the "x" represents "m", "rf", "om",
        or "orf". For non-media treatments, this argument should be set to
        `gamma_gn_dev`, which is analogous to "beta_gx_dev".

    Returns:
      The coefficient mean parameter of the treatment variable, which has
      dimension equal to the number of treatment channels..
    """
    return self.equations.calculate_beta_x(
        is_non_media=is_non_media,
        incremental_outcome_x=incremental_outcome_x,
        linear_predictor_counterfactual_difference=linear_predictor_counterfactual_difference,
        eta_x=eta_x,
        beta_gx_dev=beta_gx_dev,
    )

  # TODO: Deprecate in favor of ModelEquations.adstock_hill_media.
  def adstock_hill_media(
      self,
      media: backend.Tensor,  # pylint: disable=redefined-outer-name
      alpha: backend.Tensor,
      ec: backend.Tensor,
      slope: backend.Tensor,
      decay_functions: str | Sequence[str] = constants.GEOMETRIC_DECAY,
      n_times_output: int | None = None,
  ) -> backend.Tensor:
    """Transforms media or using Adstock and Hill functions in the desired order.

    Args:
      media: Tensor of dimensions `(n_geos, n_media_times, n_media_channels)`
        containing non-negative media execution values. Typically this is
        impressions, but it can be any metric, such as `media_spend`. Clicks are
        often used for paid search ads.
      alpha: Uniform distribution for Adstock and Hill calculations.
      ec: Shifted half-normal distribution for Adstock and Hill calculations.
      slope: Deterministic distribution for Adstock and Hill calculations.
      decay_functions: String or sequence of strings denoting the adstock decay
        function(s) for each channel. Default: 'geometric'.
      n_times_output: Number of time periods to output. This argument is
        optional when the number of time periods in `media` equals
        `self.n_media_times`, in which case `n_times_output` defaults to
        `self.n_times`.

    Returns:
      Tensor with dimensions `[..., n_geos, n_times, n_media_channels]`
      representing Adstock and Hill-transformed media.
    """
    return self.equations.adstock_hill_media(
        media=media,
        alpha=alpha,
        ec=ec,
        slope=slope,
        decay_functions=decay_functions,
        n_times_output=n_times_output,
    )

  # TODO: Deprecate in favor of ModelEquations.adstock_hill_rf.
  def adstock_hill_rf(
      self,
      reach: backend.Tensor,
      frequency: backend.Tensor,
      alpha: backend.Tensor,
      ec: backend.Tensor,
      slope: backend.Tensor,
      decay_functions: str | Sequence[str] = constants.GEOMETRIC_DECAY,
      n_times_output: int | None = None,
  ) -> backend.Tensor:
    """Transforms reach and frequency (RF) using Hill and Adstock functions.

    Args:
      reach: Tensor of dimensions `(n_geos, n_media_times, n_rf_channels)`
        containing non-negative media for reach.
      frequency: Tensor of dimensions `(n_geos, n_media_times, n_rf_channels)`
        containing non-negative media for frequency.
      alpha: Uniform distribution for Adstock and Hill calculations.
      ec: Shifted half-normal distribution for Adstock and Hill calculations.
      slope: Deterministic distribution for Adstock and Hill calculations.
      decay_functions: String or sequence of strings denoting the adstock decay
        function(s) for each channel. Default: 'geometric'.
      n_times_output: Number of time periods to output. This argument is
        optional when the number of time periods in `reach` equals
        `self.n_media_times`, in which case `n_times_output` defaults to
        `self.n_times`.

    Returns:
      Tensor with dimensions `[..., n_geos, n_times, n_rf_channels]`
      representing Hill and Adstock-transformed RF.
    """
    return self.equations.adstock_hill_rf(
        reach=reach,
        frequency=frequency,
        alpha=alpha,
        ec=ec,
        slope=slope,
        decay_functions=decay_functions,
        n_times_output=n_times_output,
    )

  def populate_cached_properties(self):
    """Eagerly activates all cached properties.

    This is useful for creating a `tf.function` computation graph with this
    Meridian object as part of a captured closure. Within the computation graph,
    internal state mutations are problematic, and so this method freezes the
    object's states before the computation graph is created.
    """
    self._model_context.populate_cached_properties()
    cls = self.__class__
    # "Freeze" all @cached_property attributes by simply accessing them (with
    # `getattr()`).
    cached_properties = [
        attr
        for attr in dir(self)
        if isinstance(getattr(cls, attr, cls), functools.cached_property)
    ]
    for attr in cached_properties:
      _ = getattr(self, attr)

  def create_inference_data_coords(
      self, n_chains: int, n_draws: int
  ) -> Mapping[str, np.ndarray | Sequence[str]]:
    """Creates data coordinates for inference data."""
    media_channel_names = (
        self.input_data.media_channel
        if self.input_data.media_channel is not None
        else np.array([])
    )
    rf_channel_names = (
        self.input_data.rf_channel
        if self.input_data.rf_channel is not None
        else np.array([])
    )
    organic_media_channel_names = (
        self.input_data.organic_media_channel
        if self.input_data.organic_media_channel is not None
        else np.array([])
    )
    organic_rf_channel_names = (
        self.input_data.organic_rf_channel
        if self.input_data.organic_rf_channel is not None
        else np.array([])
    )
    non_media_channel_names = (
        self.input_data.non_media_channel
        if self.input_data.non_media_channel is not None
        else np.array([])
    )
    control_variable_names = (
        self.input_data.control_variable
        if self.input_data.control_variable is not None
        else np.array([])
    )
    return {
        constants.CHAIN: np.arange(n_chains),
        constants.DRAW: np.arange(n_draws),
        constants.GEO: self.input_data.geo,
        constants.TIME: self.input_data.time,
        constants.MEDIA_TIME: self.input_data.media_time,
        constants.KNOTS: np.arange(self.knot_info.n_knots),
        constants.CONTROL_VARIABLE: control_variable_names,
        constants.NON_MEDIA_CHANNEL: non_media_channel_names,
        constants.MEDIA_CHANNEL: media_channel_names,
        constants.RF_CHANNEL: rf_channel_names,
        constants.ORGANIC_MEDIA_CHANNEL: organic_media_channel_names,
        constants.ORGANIC_RF_CHANNEL: organic_rf_channel_names,
    }

  def create_inference_data_dims(self) -> Mapping[str, Sequence[str]]:
    inference_dims = dict(constants.INFERENCE_DIMS)
    if self.unique_sigma_for_each_geo:
      inference_dims[constants.SIGMA] = [constants.GEO]
    else:
      inference_dims[constants.SIGMA] = []

    return {
        param: [constants.CHAIN, constants.DRAW] + list(dims)
        for param, dims in inference_dims.items()
    }

  def sample_prior(self, n_draws: int, seed: int | None = None):
    """Draws samples from the prior distributions.

    Drawn samples are merged into this model's Arviz `inference_data` property.

    Args:
      n_draws: Number of samples drawn from the prior distribution.
      seed: Used to set the seed for reproducible results. For more information,
        see [PRNGS and seeds]
        (https://github.com/tensorflow/probability/blob/main/PRNGS.md).
    """
    self.prior_sampler_callable(n_draws=n_draws, seed=seed)

  def _run_model_fitting_guardrail(self):
    """Raises an error if the model has critical EDA issues."""
    error_findings_by_type: dict[eda_outcome.EDACheckType, list[str]] = (
        collections.defaultdict(list)
    )
    for outcome in self.eda_outcomes:
      error_findings = [
          finding
          for finding in outcome.findings
          if finding.severity == eda_outcome.EDASeverity.ERROR
      ]
      if error_findings:
        error_findings_by_type[outcome.check_type].extend(
            [finding.explanation for finding in error_findings]
        )

    if error_findings_by_type:
      error_message_lines = [
          "Model has critical EDA issues. Please fix before running"
          " `sample_posterior`.\n"
      ]
      for check_type, explanations in error_findings_by_type.items():
        error_message_lines.append(f"Check type: {check_type.name}")
        for explanation in explanations:
          error_message_lines.append(f"- {explanation}")
      error_message_lines.append(
          "For further details, please refer to `Meridian.eda_outcomes`."
      )
      raise ModelFittingError("\n".join(error_message_lines))

  def sample_posterior(
      self,
      n_chains: Sequence[int] | int,
      n_adapt: int,
      n_burnin: int,
      n_keep: int,
      current_state: Mapping[str, backend.Tensor] | None = None,
      init_step_size: int | None = None,
      dual_averaging_kwargs: Mapping[str, int] | None = None,
      max_tree_depth: int = 10,
      max_energy_diff: float = 500.0,
      unrolled_leapfrog_steps: int = 1,
      parallel_iterations: int = 10,
      seed: Sequence[int] | int | None = None,
      **pins,
  ):
    """Runs Markov Chain Monte Carlo (MCMC) sampling of posterior distributions.

    For more information about the arguments, see [`windowed_adaptive_nuts`]
    (https://www.tensorflow.org/probability/api_docs/python/tfp/experimental/mcmc/windowed_adaptive_nuts).

    Drawn samples are merged into this model's Arviz `inference_data` property.

    Args:
      n_chains: Number of MCMC chains. Given a sequence of integers,
        `windowed_adaptive_nuts` will be called once for each element. The
        `n_chains` argument of each `windowed_adaptive_nuts` call will be equal
        to the respective integer element. Using a list of integers, one can
        split the chains of a `windowed_adaptive_nuts` call into multiple calls
        with fewer chains per call. This can reduce memory usage. This might
        require an increased number of adaptation steps for convergence, as the
        optimization is occurring across fewer chains per sampling call.
      n_adapt: Number of adaptation draws per chain.
      n_burnin: Number of burn-in draws per chain. Burn-in draws occur after
        adaptation draws and before the kept draws.
      n_keep: Integer number of draws per chain to keep for inference.
      current_state: Optional structure of tensors at which to initialize
        sampling. Use the same shape and structure as
        `model.experimental_pin(**pins).sample(n_chains)`.
      init_step_size: Optional integer determining where to initialize the step
        size for the leapfrog integrator. The structure must broadcast with
        `current_state`. For example, if the initial state is:  ``` { 'a':
        tf.zeros(n_chains), 'b': tf.zeros([n_chains, n_features]), } ```  then
        any of `1.`, `{'a': 1., 'b': 1.}`, or `{'a': tf.ones(n_chains), 'b':
        tf.ones([n_chains, n_features])}` will work. Defaults to the dimension
        of the log density to the ¼ power.
      dual_averaging_kwargs: Optional dict keyword arguments to pass to
        `tfp.mcmc.DualAveragingStepSizeAdaptation`. By default, a
        `target_accept_prob` of `0.85` is set, acceptance probabilities across
        chains are reduced using a harmonic mean, and the class defaults are
        used otherwise.
      max_tree_depth: Maximum depth of the tree implicitly built by NUTS. The
        maximum number of leapfrog steps is bounded by `2**max_tree_depth`, for
        example, the number of nodes in a binary tree `max_tree_depth` nodes
        deep. The default setting of `10` takes up to 1024 leapfrog steps.
      max_energy_diff: Scalar threshold of energy differences at each leapfrog,
        divergence samples are defined as leapfrog steps that exceed this
        threshold. Default is `1000`.
      unrolled_leapfrog_steps: The number of leapfrogs to unroll per tree
        expansion step. Applies a direct linear multiplier to the maximum
        trajectory length implied by `max_tree_depth`. Defaults is `1`.
      parallel_iterations: Number of iterations allowed to run in parallel. Must
        be a positive integer. For more information, see `tf.while_loop`.
      seed: An `int32[2]` Tensor or a Python list or tuple of 2 `int`s, which
        will be treated as stateless seeds; or a Python `int` or `None`, which
        will be treated as stateful seeds. See [tfp.random.sanitize_seed]
        (https://www.tensorflow.org/probability/api_docs/python/tfp/random/sanitize_seed).
      **pins: These are used to condition the provided joint distribution, and
        are passed directly to `joint_dist.experimental_pin(**pins)`.

    Throws:
      MCMCOOMError: If the model is out of memory. Try reducing `n_keep` or pass
        a list of integers as `n_chains` to sample chains serially. For more
        information, see
        [ResourceExhaustedError when running Meridian.sample_posterior]
        (https://developers.google.com/meridian/docs/post-modeling/model-debugging#gpu-oom-error).
    """
    self._run_model_fitting_guardrail()

    self.posterior_sampler_callable(
        n_chains=n_chains,
        n_adapt=n_adapt,
        n_burnin=n_burnin,
        n_keep=n_keep,
        current_state=current_state,
        init_step_size=init_step_size,
        dual_averaging_kwargs=dual_averaging_kwargs,
        max_tree_depth=max_tree_depth,
        max_energy_diff=max_energy_diff,
        unrolled_leapfrog_steps=unrolled_leapfrog_steps,
        parallel_iterations=parallel_iterations,
        seed=seed,
        **pins,
    )


def save_mmm(mmm: Meridian, file_path: str):
  """Save the model object to a `pickle` file path.

  WARNING: There is no guarantee for future compatibility of the binary file
  output of this function. We recommend using `load_mmm()` with the same
  version of the library that was used to save the model.

  Args:
    mmm: Model object to save.
    file_path: File path to save a pickled model object.
  """
  if not os.path.exists(os.path.dirname(file_path)):
    os.makedirs(os.path.dirname(file_path))

  with open(file_path, "wb") as f:
    joblib.dump(mmm, f)


def load_mmm(file_path: str) -> Meridian:
  """Load the model object from a `pickle` file path.

  WARNING: There is no guarantee for backward compatibility of the binary file
  input of this function. We recommend using `load_mmm()` with the same
  version of the library that was used to save the model's pickled file.

  Args:
    file_path: File path to load a pickled model object from.

  Returns:
    mmm: Model object loaded from the file path.

  Raises:
      FileNotFoundError: If `file_path` does not exist.
  """
  try:
    with open(file_path, "rb") as f:
      mmm = joblib.load(f)
    return mmm
  except FileNotFoundError:
    raise FileNotFoundError(f"No such file or directory: {file_path}") from None



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configurations for the Model Quality Checks."""

import dataclasses


@dataclasses.dataclass(frozen=True)
class BaseConfig:
  """Base class for all check configurations."""


@dataclasses.dataclass(frozen=True)
class ConvergenceConfig(BaseConfig):
  """Configuration for the Convergence Check.

  Attributes:
    convergence_threshold: The threshold for the R-hat statistic to determine if
      the model has converged. R-hat values below this are considered converged.
    not_fully_convergence_threshold: The threshold for the R-hat statistic to
      determine if the model is not fully converged but potentially acceptable.
      R-hat values between `convergence_threshold` and this value are considered
      not fully converged. R-hat values above this threshold are considered not
      converged.
  """

  convergence_threshold: float = 1.2
  not_fully_convergence_threshold: float = 10.0


@dataclasses.dataclass(frozen=True)
class ROIConsistencyConfig(BaseConfig):
  """Configuration for the ROI Consistency Check.

  This check verifies if the posterior median of the ROI falls within a
  reasonable range of the prior distribution.

  Attributes:
    prior_lower_quantile: The lower quantile of the ROI prior distribution to
      define the lower bound of the reasonable range.
    prior_upper_quantile: The upper quantile of the ROI prior distribution to
      define the upper bound of the reasonable range.
  """

  prior_lower_quantile: float = 0.01
  prior_upper_quantile: float = 0.99


@dataclasses.dataclass(frozen=True)
class BaselineConfig(BaseConfig):
  """Configuration for the Baseline Check.

  This check warns if there is a high probability of a negative baseline.

  Attributes:
    negative_baseline_prob_review_threshold: Probability threshold for a
      review. If the probability of a negative baseline is above this value, a
      review is issued.
    negative_baseline_prob_fail_threshold: Probability threshold for a failure.
      If the probability of a negative baseline is above this value, the check
      fails.
  """

  negative_baseline_prob_review_threshold: float = 0.2
  negative_baseline_prob_fail_threshold: float = 0.8


@dataclasses.dataclass(frozen=True)
class BayesianPPPConfig(BaseConfig):
  """Configuration for the Bayesian Posterior Predictive P-value Check.

  Attributes:
    ppp_threshold: P-value threshold for posterior predictive check.
  """

  ppp_threshold: float = 0.05


@dataclasses.dataclass(frozen=True)
class GoodnessOfFitConfig(BaseConfig):
  """An empty config for the Goodness of Fit Check."""


@dataclasses.dataclass(frozen=True)
class PriorPosteriorShiftConfig(BaseConfig):
  """Configuration for the Prior-Posterior Shift Check.

  Attributes:
    n_bootstraps: Number of bootstrap samples to use for calculating posterior
      statistics.
    alpha: Significance level for detecting a shift between prior and posterior
      distributions.
    seed: Random seed for reproducibility of bootstrap sampling.
  """

  n_bootstraps: int = 1000
  alpha: float = 0.05
  seed: int = 42



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper classes for generating date intervals for various time buckets."""

import abc
from collections.abc import Iterator, Sequence
import datetime
from typing import TypeAlias


__all__ = [
    "DateRangeBucketer",
    "MonthlyDateRangeGenerator",
    "QuarterlyDateRangeGenerator",
    "YearlyDateRangeGenerator",
]


DateInterval: TypeAlias = tuple[datetime.date, datetime.date]


class DateRangeBucketer(abc.ABC):
  """Generates `DateInterval` protos over a range of dates."""

  def __init__(
      self,
      input_dates: Sequence[datetime.date],
  ):
    """Initializes the DateRangeBucketer with a sequence of dates.

    Args:
      input_dates: A sequence of `datetime.date` objects representing the range
        of dates to generate intervals for.
    """
    if not all(
        input_dates[i] < input_dates[i + 1] for i in range(len(input_dates) - 1)
    ):
      raise ValueError("`input_dates` must be strictly ascending dates.")

    self._input_dates = input_dates

  @abc.abstractmethod
  def generate_date_intervals(self) -> Iterator[DateInterval]:
    """Generates `DateInterval` protos for the class's input dates.

    Each interval represents a month, quarter, or year, depending on the
    instance of this class. An interval is excluded if the start date is not the
    first available date (in `self._input_dates`) for the time bucket. The last
    interval in `self._input_dates` is excluded in all cases.

    Returns:
      An iterator over generated `TimeInterval`s.
    """
    raise NotImplementedError()


class MonthlyDateRangeGenerator(DateRangeBucketer):
  """Generates monthly date intervals."""

  def generate_date_intervals(self) -> Iterator[DateInterval]:
    start_date = self._input_dates[0]

    for date in self._input_dates:
      if date.month != start_date.month:
        if start_date.day <= 7:
          yield (start_date, date)

        start_date = date


class QuarterlyDateRangeGenerator(DateRangeBucketer):
  """Generates quarterly date intervals."""

  def generate_date_intervals(self) -> Iterator[DateInterval]:
    start_date = self._input_dates[0]
    for date in self._input_dates:
      start_date_quarter_number = (start_date.month - 1) // 3 + 1
      current_date_quarter_number = (date.month - 1) // 3 + 1

      if start_date_quarter_number != current_date_quarter_number:
        # The interval is only included if the start date is the first date of
        # the quarter that's present in `self._input_dates`. We can detect this
        # date by checking whether it's in the first month of the quarter and
        # falls in the first seven days of the month.
        if (
            start_date.day <= 7
            and start_date.month == ((start_date_quarter_number - 1) * 3) + 1
        ):
          yield (start_date, date)

        start_date = date


class YearlyDateRangeGenerator(DateRangeBucketer):
  """Generates yearly date intervals."""

  def generate_date_intervals(self) -> Iterator[DateInterval]:
    start_date = self._input_dates[0]

    for date in self._input_dates:
      if date.year != start_date.year:
        if start_date.day <= 7 and start_date.month == 1:
          yield (start_date, date)

        start_date = date



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meridian analysis API for trained models."""

from meridian.analysis import analyzer
from meridian.analysis import optimizer
from meridian.analysis import review
from meridian.analysis import summarizer
from meridian.analysis import visualizer
from meridian.templates import formatter
